# Copyright (C) 2022-2025 Jørgen S. Dokken, Henrik N. T. Finsberg and
# Paul T. Kühner
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Tools to extract data from Gmsh models."""

import typing
from collections.abc import Callable
from pathlib import Path

from mpi4py import MPI as _MPI

import numpy as np
import numpy.typing as npt

import basix
import basix.ufl
import ufl
from dolfinx import cpp as _cpp
from dolfinx import default_real_type
from dolfinx.cpp.graph import AdjacencyList_int32 as _AdjacencyList_int32
from dolfinx.graph import AdjacencyList, adjacencylist
from dolfinx.io.utils import distribute_entity_data
from dolfinx.mesh import CellType, Mesh, MeshTags, create_mesh, meshtags_from_entities

__all__ = [
    "MeshData",
    "PhysicalGroup",
    "TopologyDict",
    "cell_perm_array",
    "extract_geometry",
    "extract_topology_and_markers",
    "model_to_mesh",
    "read_from_msh",
    "ufl_mesh",
]


class TopologyDict(typing.TypedDict):
    """TopologyDict is a TypedDict for storing the topology of the marked
    cell.

    Args:
        topology: 2D array containing the topology of the marked cell.
        cell_data: List with the corresponding markers.

    Note:
        The TypedDict is only used for type hinting, and does not
        enforce the structure of the dictionary, but rather provides
        a hint to the user and the type checker.
    """

    topology: npt.NDArray[
        np.integer
    ]  # cell->node connectivity, shape (num_cells, num_nodes_per_cell)
    cell_data: npt.NDArray[np.integer]  # cell tags, shape (num_cells,)
    entity_tags: npt.NDArray[np.integer]  # GMSH cell indices, shape (num_cells,)


# Map from Gmsh cell type identifier (integer) to DOLFINx cell type and
# degree https://gmsh.info//doc/texinfo/gmsh.html#MSH-file-format
_gmsh_to_cells = {
    1: ("interval", 1),
    2: ("triangle", 1),
    3: ("quadrilateral", 1),
    4: ("tetrahedron", 1),
    5: ("hexahedron", 1),
    8: ("interval", 2),
    9: ("triangle", 2),
    10: ("quadrilateral", 2),
    11: ("tetrahedron", 2),
    12: ("hexahedron", 2),
    15: ("point", 0),
    21: ("triangle", 3),
    26: ("interval", 3),
    29: ("tetrahedron", 3),
    36: ("quadrilateral", 3),
    92: ("hexahedron", 3),
}


class PhysicalGroup(typing.NamedTuple):
    """Physical group info.

    Args:
        dim: dimension of the physical group
        tag: tag of the physical group
    """

    dim: int
    tag: int


class MeshData(typing.NamedTuple):
    """Data for representing a mesh and associated tags.

    Args:
        mesh: Mesh.
        cell_tags: MeshTags for cells.
        facet_tags: MeshTags for facets (codim 1).
        ridge_tags: MeshTags for ridges (codim 2).
        peak_tags: MeshTags for peaks (codim 3).
        physical_groups: Physical groups in the mesh, where the key
            is the physical name and the value is a tuple with the
            dimension and tag.
    """

    mesh: Mesh
    cell_tags: MeshTags | None
    facet_tags: MeshTags | None
    ridge_tags: MeshTags | None
    peak_tags: MeshTags | None
    physical_groups: dict[str, PhysicalGroup]


def ufl_mesh(gmsh_cell: int, gdim: int, dtype: npt.DTypeLike) -> ufl.Mesh:
    """Create a UFL mesh from a Gmsh cell identifier and geometric
    dimension.

    See https://gmsh.info//doc/texinfo/gmsh.html#MSH-file-format.

    Args:
        gmsh_cell: Gmsh cell identifier.
        gdim: Geometric dimension of the mesh.

    Returns:
        UFL Mesh using Lagrange elements (equispaced) of the
        corresponding DOLFINx cell.

    """
    try:
        shape, degree = _gmsh_to_cells[gmsh_cell]
    except KeyError as e:
        print(f"Unknown cell type {gmsh_cell}.")
        raise e
    cell = ufl.Cell(shape)
    element = basix.ufl.element(
        basix.ElementFamily.P,
        cell.cellname(),
        degree,
        basix.LagrangeVariant.equispaced,
        shape=(gdim,),
        dtype=dtype,  # type: ignore[arg-type]
    )
    return ufl.Mesh(element)


def cell_perm_array(cell_type: CellType, num_nodes: int) -> list[int]:
    """The permutation array for permuting Gmsh ordering to DOLFINx
    ordering.

    Args:
        cell_type: DOLFINx cell type.
        num_nodes: Number of nodes in the cell.

    Returns:
        An array ``p`` such that ``a_dolfinx[i] = a_gmsh[p[i]]``.

    """
    return _cpp.io.perm_gmsh(cell_type, num_nodes)


def extract_topology_and_markers(
    model, name: str | None = None
) -> tuple[dict[int, TopologyDict], dict[str, PhysicalGroup]]:
    """Extract all entities tagged with a physical marker in the gmsh
    model.

    Returns a nested dictionary where the first key is the gmsh MSH
    element type integer. Each element type present in the model
    contains the cell topology of the elements and corresponding
    markers.

    Args:
        model: Gmsh model.
        name: Name of the gmsh model. If not set the current
            model will be used.

    Returns:
        A tuple ``(topologies, physical_groups)``, where ``topologies`` is
        a nested dictionary where each key corresponds to a gmsh cell type.
        Each cell type found in the mesh has a 2D array containing the
        topology of the marked cell and a list with the corresponding
        markers. ``physical_groups`` is a dictionary where the key is the
        physical name and the value is a tuple with the dimension and tag.

    """
    if name is not None:
        model.setCurrent(name)

    # Get the physical groups from gmsh in the form [(dim1, tag1),
    # (dim1, tag2), (dim2, tag3),...]
    phys_grps = model.getPhysicalGroups()
    topologies: dict[int, TopologyDict] = {}
    # Create a dictionary with the physical groups where the key is the
    # physical name and the value is a tuple with the dimension and tag
    physical_groups: dict[str, PhysicalGroup] = {}
    for dim, tag in phys_grps:
        # Get the entities of dimension `dim`, dim=0 -> Points, dim=1 -
        # >Lines, dim=2 -> Triangles/Quadrilaterals, etc.
        entities = model.getEntitiesForPhysicalGroup(dim, tag)
        for entity in entities:
            # Get cell type, list of cells with given tag and topology
            # of tagged cells NOTE: Assumes that each entity only have
            # one cell-type,
            # i.e. facets of prisms and pyramid meshes are not supported
            (entity_types, entity_tags, entity_topologies) = model.mesh.getElements(dim, tag=entity)

            if len(entity_types) > 1:
                raise RuntimeError(
                    f"Unsupported mesh with multiple cell types {entity_types} for entity {entity}"
                )
            elif len(entity_types) == 0:
                continue

            # Determine number of local nodes per element to create the
            # topology of the elements
            properties = model.mesh.getElementProperties(entity_types[0])
            name, dim, _, num_nodes, _, _ = properties

            # Array of shape (num_elements,num_nodes_per_element)
            # containing the topology of the elements on this entity.
            # NOTE: Gmsh indexing starts with one, we therefore subtract
            # 1 from each node to use zero-based numbering
            topology = entity_topologies[0].reshape(-1, num_nodes) - 1

            # Create marker array of length of number of tagged cells
            marker = np.full_like(entity_tags[0], tag)

            # Keep track of entity tags to check tag consistency
            entity_tags = entity_tags[0]

            # Group element topology and markers of the same entity type
            entity_type = entity_types[0]
            if entity_type in topologies.keys():
                topologies[entity_type]["topology"] = np.concatenate(
                    (topologies[entity_type]["topology"], topology), axis=0
                )
                topologies[entity_type]["cell_data"] = np.hstack(
                    [topologies[entity_type]["cell_data"], marker]
                )
                topologies[entity_type]["entity_tags"] = np.hstack(
                    [topologies[entity_type]["entity_tags"], entity_tags]
                )
            else:
                topologies[entity_type] = {
                    "topology": topology,
                    "cell_data": marker,
                    "entity_tags": entity_tags,
                }
        physical_groups[model.getPhysicalName(dim, tag)] = PhysicalGroup(dim, tag)

    return topologies, physical_groups


def extract_geometry(model, name: str | None = None) -> npt.NDArray[np.float64]:
    """Extract the mesh geometry from a Gmsh model.

    Returns an array of shape ``(num_nodes, 3)``, where the i-th row
    corresponds to the i-th node in the mesh.

    Args:
        model: Gmsh model
        name: Name of the Gmsh model. If not set the current
            model will be used.

    Returns:
        The mesh geometry as an array of shape ``(num_nodes, 3)``.

    """
    if name is not None:
        model.setCurrent(name)

    # Get the unique tag and coordinates for nodes in mesh
    indices, points, _ = model.mesh.getNodes()
    points = points.reshape(-1, 3)

    # Gmsh indices starts at 1. We therefore subtract one to use
    # zero-based numbering
    indices -= 1

    # In some cases, Gmsh does not return the points in the same
    # order as their unique node index. We therefore sort nodes in
    # geometry according to the unique index
    perm_sort = np.argsort(indices)
    assert np.all(indices[perm_sort] == np.arange(len(indices)))
    return points[perm_sort]


def model_to_mesh(
    model,
    comm: _MPI.Comm,
    rank: int,
    gdim: int = 3,
    partitioner: Callable[[_MPI.Comm, int, int, _AdjacencyList_int32], _AdjacencyList_int32]
    | None = None,
    dtype=default_real_type,
) -> MeshData:
    """Create a Mesh from a Gmsh model.

    Creates a :class:`dolfinx.mesh.Mesh` from the physical entities of
    the highest topological dimension in the Gmsh model. In parallel,
    the gmsh model is processed on one MPI rank, and the
    :class:`dolfinx.mesh.Mesh` is distributed across ranks.

    Args:
        model: Gmsh model.
        comm: MPI communicator to use for mesh creation.
        rank: MPI rank that the Gmsh model is initialized on.
        gdim: Geometrical dimension of the mesh.
        partitioner: Function that computes the parallel
            distribution of cells across MPI ranks.

    Returns:
        MeshData with mesh and tags of corresponding entities by
        codimension. Codimension 0 is the cell tags, codimension 1 is the
        facet tags, codimension 2 is the ridge tags and codimension 3 is
        the peak tags as well as a lookup table from the physical groups by
        name to integer.

    Note:
        For performance, this function should only be called once for
        large problems. For reuse, it is recommended to save the mesh
        and corresponding tags using :class:`dolfinx.io.XDMFFile` after
        creation for efficient access.
    """
    valid_mesh = None
    if comm.rank == rank:
        assert model is not None, "Gmsh model is None on rank responsible for mesh creation."
        # Get mesh geometry and mesh topology for each element
        x = extract_geometry(model)
        topologies, physical_groups = extract_topology_and_markers(model)
        valid_mesh = len(physical_groups) > 0

        # Extract Gmsh entity (cell) id, topological dimension and number
        # of nodes which is used to create an appropriate coordinate
        # element, and seperate higher topological entities from lower
        # topological entities (e.g. facets, ridges and peaks).
        num_unique_entities = len(topologies.keys())
        element_ids = np.zeros(num_unique_entities, dtype=np.int32)
        entity_tdim = np.zeros(num_unique_entities, dtype=np.int32)
        num_nodes_per_element = np.zeros(num_unique_entities, dtype=np.int32)
        for i, element in enumerate(topologies.keys()):
            _, dim, _, num_nodes, _, _ = model.mesh.getElementProperties(element)
            element_ids[i] = element
            entity_tdim[i] = dim
            num_nodes_per_element[i] = num_nodes

        # Broadcast information to all other ranks
        entity_tdim, element_ids, num_nodes_per_element = comm.bcast(
            (entity_tdim, element_ids, num_nodes_per_element), root=rank
        )
    else:
        entity_tdim, element_ids, num_nodes_per_element = comm.bcast((None, None, None), root=rank)

    valid_mesh = comm.bcast(valid_mesh, root=rank)
    if not valid_mesh:
        raise RuntimeError("No 'physical groups' in gmsh mesh. Cannot continue.")

    # Sort elements by descending dimension
    assert len(np.unique(entity_tdim)) == len(entity_tdim)
    perm_sort = np.argsort(entity_tdim)[::-1]

    # Extract position of the highest topological entity and its
    # topological dimension
    cell_position = perm_sort[0]
    tdim = int(entity_tdim[cell_position])

    # Check that all cells are tagged once
    error_msg = ""
    if comm.rank == rank:
        _elementTypes, _elementTags, _nodeTags = model.mesh.getElements(dim=tdim, tag=-1)
        _elementType_dim = _elementTypes[0]
        if _elementType_dim not in topologies.keys():
            error_msg = "All cells are expected to be tagged once; none found"

        num_cells_tagged = len(topologies[_elementType_dim]["entity_tags"])
        if (num_cells := len(_elementTags[0])) != num_cells_tagged:
            error_msg = (
                "All cells are expected to be tagged once;"
                + f"found: {num_cells_tagged}, expected: {num_cells}"
            )
        num_cells_tagged_once = len(np.unique(topologies[_elementType_dim]["entity_tags"]))
        if num_cells_tagged != num_cells_tagged_once:
            error_msg = "All cells are expected to be tagged once, found duplicates"

    error_msg = comm.bcast(error_msg, root=rank)
    if error_msg != "":
        raise RuntimeError(error_msg)

    # Extract entity -> node connectivity for all cells and sub-entities
    # marked in the GMSH model
    meshtags: dict[int, tuple[npt.NDArray[np.int64], npt.NDArray[np.int32]]] = {}
    for position in perm_sort:
        codim = tdim - entity_tdim[position]
        if comm.rank == rank:
            gmsh_entity_id = element_ids[position]
            marked_entities = np.asarray(topologies[gmsh_entity_id]["topology"], dtype=np.int64)
            entity_values = np.asarray(topologies[gmsh_entity_id]["cell_data"], dtype=np.int32)
            meshtags[codim] = (marked_entities, entity_values)
        else:
            # Any other process than input rank does not have any entities
            marked_entities = np.empty((0, num_nodes_per_element[position]), dtype=np.int32)
            entity_values = np.empty((0,), dtype=np.int32)
            meshtags[codim] = (marked_entities, entity_values)

    # Create a UFL Mesh object for the GMSH element with the highest
    # topoligcal dimension
    ufl_domain = ufl_mesh(element_ids[cell_position], gdim, dtype=dtype)

    # Get cell->node connectivity and  permute to FEniCS ordering
    num_nodes = num_nodes_per_element[cell_position]
    gmsh_cell_perm = cell_perm_array(_cpp.mesh.to_type(str(ufl_domain.ufl_cell())), num_nodes)
    cell_connectivity = meshtags[0][0][:, gmsh_cell_perm].copy()

    # Create a distributed mesh, where mesh nodes are only destributed from
    # the input rank
    if comm.rank != rank:
        x = np.empty([0, gdim], dtype=dtype)  # No nodes on other than root rank
    mesh = create_mesh(
        comm, cell_connectivity, ufl_domain, x[:, :gdim].astype(dtype, copy=False), partitioner
    )
    assert tdim == mesh.topology.dim, (
        f"{mesh.topology.dim=} does not match Gmsh model dimension {tdim}"
    )

    # Create MeshTags for all sub entities
    topology = mesh.topology
    codim_to_name = {0: "cell", 1: "facet", 2: "ridge", 3: "peak"}
    dolfinx_meshtags: dict[str, MeshTags | None] = {}
    for codim in [0, 1, 2, 3]:
        key = f"{codim_to_name[codim]}_tags"
        if (
            codim == 1 and topology.cell_type == CellType.prism
        ) or topology.cell_type == CellType.pyramid:
            raise RuntimeError(f"Unsupported facet tag for type {topology.cell_type}")

        meshtag_data = meshtags.get(codim, None)
        if meshtag_data is None:
            dolfinx_meshtags[key] = None
            continue

        # Distribute entity data [[e0_v0, e0_v1, ...], [e1_v0, e1_v1, ...],
        # ...] which is made in global input indices to local indices on
        # the owning process
        (marked_entities, entity_values) = meshtag_data
        local_entities, local_values = distribute_entity_data(
            mesh, tdim - codim, marked_entities, entity_values
        )
        # Create MeshTags object from the local entities
        mesh.topology.create_connectivity(tdim - codim, tdim)
        adj = adjacencylist(local_entities)
        et = meshtags_from_entities(
            mesh, tdim - codim, adj, local_values.astype(np.int32, copy=False)
        )
        et.name = key
        dolfinx_meshtags[key] = et

    # Broadcast physical groups (string to integer mapping) to all ranks
    if comm.rank == rank:
        physical_groups = comm.bcast(physical_groups, root=rank)
    else:
        physical_groups = comm.bcast(None, root=rank)

    return MeshData(mesh, physical_groups=physical_groups, **dolfinx_meshtags)


def read_from_msh(
    filename: str | Path,
    comm: _MPI.Comm,
    rank: int = 0,
    gdim: int = 3,
    partitioner: Callable[[_MPI.Comm, int, int, AdjacencyList], _AdjacencyList_int32] | None = None,
) -> MeshData:
    """Read a Gmsh .msh file and return a :class:`dolfinx.mesh.Mesh` and
    cell facet markers.

    Note:
        This function requires the Gmsh Python module.

    Args:
        filename: Name of ``.msh`` file.
        comm: MPI communicator to create the mesh on.
        rank: Rank of ``comm`` responsible for reading the ``.msh``
            file.
        gdim: Geometric dimension of the mesh

    Returns:
        Meshdata with mesh, cell tags, facet tags, edge tags,
        vertex tags and physical groups.

    """
    try:
        import gmsh
    except ModuleNotFoundError:
        # Python 3.11+ adds the add_note method to exceptions
        # e.add_note("Gmsh must be installed to import dolfinx.io.gmsh")
        raise ModuleNotFoundError(
            "No module named 'gmsh': dolfinx.io.gmsh.read_from_msh requires Gmsh.", name="gmsh"
        )

    if comm.rank == rank:
        gmsh.initialize()
        gmsh.model.add("Mesh from file")
        gmsh.merge(str(filename))
        msh = model_to_mesh(gmsh.model, comm, rank, gdim=gdim, partitioner=partitioner)
        gmsh.finalize()
        return msh
    else:
        return model_to_mesh(gmsh.model, comm, rank, gdim=gdim, partitioner=partitioner)



"""Unit tests for assembly with Numba and CFFI kernels."""

# Copyright (C) 2018-2014 Chris N. Richardson, Michal Habera and Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import os
import sys

from mpi4py import MPI

import numpy as np
import pytest

import dolfinx
import ffcx.codegeneration.utils
from dolfinx import cpp as _cpp
from dolfinx import fem, la
from dolfinx.common import list_timings
from dolfinx.fem import Form, Function, IntegralType, form_cpp_class, functionspace
from dolfinx.mesh import create_unit_square

numba = pytest.importorskip("numba")
ufcx_signature = ffcx.codegeneration.utils.numba_ufcx_kernel_signature

sys.path.append(os.getcwd())


def tabulate_rank2(dtype, xdtype):
    @numba.cfunc(ufcx_signature(dtype, xdtype), nopython=True)
    def tabulate(A_, w_, c_, coords_, entity_local_index, cell_orientation, custom_data):
        A = numba.carray(A_, (3, 3), dtype=dtype)
        coordinate_dofs = numba.carray(coords_, (3, 3), dtype=xdtype)

        # Ke=∫Ωe BTe Be dΩ
        x0, y0 = coordinate_dofs[0, :2]
        x1, y1 = coordinate_dofs[1, :2]
        x2, y2 = coordinate_dofs[2, :2]

        # 2x Element area Ae
        Ae = abs((x0 - x1) * (y2 - y1) - (y0 - y1) * (x2 - x1))
        B = np.array([y1 - y2, y2 - y0, y0 - y1, x2 - x1, x0 - x2, x1 - x0], dtype=dtype).reshape(
            2, 3
        )
        A[:, :] = np.dot(B.T, B) / (2 * Ae)

    return tabulate


def tabulate_rank1(dtype, xdtype):
    @numba.cfunc(ufcx_signature(dtype, xdtype), nopython=True)
    def tabulate(b_, w_, c_, coords_, local_index, orientation, custom_data):
        b = numba.carray(b_, (3), dtype=dtype)
        coordinate_dofs = numba.carray(coords_, (3, 3), dtype=xdtype)
        x0, y0 = coordinate_dofs[0, :2]
        x1, y1 = coordinate_dofs[1, :2]
        x2, y2 = coordinate_dofs[2, :2]

        # 2x Element area Ae
        Ae = abs((x0 - x1) * (y2 - y1) - (y0 - y1) * (x2 - x1))
        b[:] = Ae / 6.0

    return tabulate


def tabulate_rank1_coeff(dtype, xdtype):
    @numba.cfunc(ufcx_signature(dtype, xdtype), nopython=True)
    def tabulate(b_, w_, c_, coords_, local_index, orientation, custom_data):
        b = numba.carray(b_, (3), dtype=dtype)
        w = numba.carray(w_, (1), dtype=dtype)
        coordinate_dofs = numba.carray(coords_, (3, 3), dtype=xdtype)
        x0, y0 = coordinate_dofs[0, :2]
        x1, y1 = coordinate_dofs[1, :2]
        x2, y2 = coordinate_dofs[2, :2]

        # 2x Element area Ae
        Ae = abs((x0 - x1) * (y2 - y1) - (y0 - y1) * (x2 - x1))
        b[:] = w[0] * Ae / 6.0

    return tabulate


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
def test_numba_assembly(dtype):
    xdtype = np.real(dtype(0)).dtype
    k2 = tabulate_rank2(dtype, xdtype)
    k1 = tabulate_rank1(dtype, xdtype)
    mesh = create_unit_square(MPI.COMM_WORLD, 13, 13, dtype=xdtype)
    V = functionspace(mesh, ("Lagrange", 1))
    cells = np.arange(mesh.topology.index_map(mesh.topology.dim).size_local, dtype=np.int32)
    active_coeffs = np.array([], dtype=np.int8)
    integrals = {
        IntegralType.cell: [
            (0, k2.address, cells, active_coeffs),
            (1, k2.address, np.arange(0), active_coeffs),
            (2, k2.address, np.arange(0), active_coeffs),
        ]
    }
    formtype = form_cpp_class(dtype)
    a = Form(
        formtype(
            [V._cpp_object, V._cpp_object], integrals, [], [], False, [], mesh=mesh._cpp_object
        )
    )
    integrals = {IntegralType.cell: [(0, k1.address, cells, active_coeffs)]}
    L = Form(formtype([V._cpp_object], integrals, [], [], False, [], mesh=mesh._cpp_object))

    A = dolfinx.fem.assemble_matrix(a)
    A.scatter_reverse()
    b = dolfinx.fem.assemble_vector(L)
    b.scatter_reverse(dolfinx.la.InsertMode.add)

    Anorm = np.sqrt(A.squared_norm())
    bnorm = la.norm(b)
    assert np.isclose(Anorm, 56.124860801609124)
    assert np.isclose(bnorm, 0.0739710713711999)

    list_timings(MPI.COMM_WORLD)


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
def test_coefficient(dtype):
    xdtype = np.real(dtype(0)).dtype
    k1 = tabulate_rank1_coeff(dtype, xdtype)

    mesh = create_unit_square(MPI.COMM_WORLD, 13, 13, dtype=xdtype)
    V = functionspace(mesh, ("Lagrange", 1))
    DG0 = functionspace(mesh, ("DG", 0))
    vals = Function(DG0, dtype=dtype)
    vals.x.array[: vals.x.index_map.size_local] = 2

    tdim = mesh.topology.dim
    num_cells = mesh.topology.index_map(tdim).size_local + mesh.topology.index_map(tdim).num_ghosts
    active_coeffs = np.array([0], dtype=np.int8)
    integrals = {
        IntegralType.cell: [(0, k1.address, np.arange(num_cells, dtype=np.int32), active_coeffs)]
    }
    formtype = form_cpp_class(dtype)
    L = Form(
        formtype(
            [V._cpp_object], integrals, [vals._cpp_object], [], False, [], mesh=mesh._cpp_object
        )
    )

    b = dolfinx.fem.assemble_vector(L)
    b.scatter_reverse(la.InsertMode.add)
    bnorm = la.norm(b)
    assert np.isclose(bnorm, 2.0 * 0.0739710713711999)


@pytest.mark.skip_in_parallel
def test_cffi_assembly():
    pytest.importorskip("cffi")
    mesh = create_unit_square(MPI.COMM_WORLD, 13, 13, dtype=np.float64)
    V = functionspace(mesh, ("Lagrange", 1))
    if mesh.comm.rank == 0:
        from cffi import FFI

        ffibuilder = FFI()
        ffibuilder.set_source(
            "_cffi_kernelA",
            r"""
        #include <math.h>
        void tabulate_tensor_poissonA(double* restrict A, const double* w,
                                    const double* c,
                                    const double* restrict coordinate_dofs,
                                    const int* entity_local_index,
                                    const int* cell_orientation,
                                    void* custom_data)
        {
        // Precomputed values of basis functions and precomputations
        // FE* dimensions: [entities][points][dofs]
        // PI* dimensions: [entities][dofs][dofs] or [entities][dofs]
        // PM* dimensions: [entities][dofs][dofs]
        static const double FE3_C0_D01_Q1[1][1][2] = { { { -1.0, 1.0 } } };
        // Unstructured piecewise computations
        const double J_c0 = coordinate_dofs[0] * FE3_C0_D01_Q1[0][0][0]
          + coordinate_dofs[3] * FE3_C0_D01_Q1[0][0][1];
        const double J_c3 = coordinate_dofs[1] * FE3_C0_D01_Q1[0][0][0]
          + coordinate_dofs[7] * FE3_C0_D01_Q1[0][0][1];
        const double J_c1 = coordinate_dofs[0] * FE3_C0_D01_Q1[0][0][0]
          + coordinate_dofs[6] * FE3_C0_D01_Q1[0][0][1];
        const double J_c2 = coordinate_dofs[1] * FE3_C0_D01_Q1[0][0][0]
          + coordinate_dofs[4] * FE3_C0_D01_Q1[0][0][1];
        double sp[20];
        sp[0] = J_c0 * J_c3;
        sp[1] = J_c1 * J_c2;
        sp[2] = sp[0] + -1 * sp[1];
        sp[3] = J_c0 / sp[2];
        sp[4] = -1 * J_c1 / sp[2];
        sp[5] = sp[3] * sp[3];
        sp[6] = sp[3] * sp[4];
        sp[7] = sp[4] * sp[4];
        sp[8] = J_c3 / sp[2];
        sp[9] = -1 * J_c2 / sp[2];
        sp[10] = sp[9] * sp[9];
        sp[11] = sp[8] * sp[9];
        sp[12] = sp[8] * sp[8];
        sp[13] = sp[5] + sp[10];
        sp[14] = sp[6] + sp[11];
        sp[15] = sp[12] + sp[7];
        sp[16] = fabs(sp[2]);
        sp[17] = sp[13] * sp[16];
        sp[18] = sp[14] * sp[16];
        sp[19] = sp[15] * sp[16];
        // UFLACS block mode: preintegrated
        A[0] = 0.5 * sp[19] + 0.5 * sp[18] + 0.5 * sp[18] + 0.5 * sp[17];
        A[1] = -0.5 * sp[19] + -0.5 * sp[18];
        A[2] = -0.5 * sp[18] + -0.5 * sp[17];
        A[3] = -0.5 * sp[19] + -0.5 * sp[18];
        A[4] = 0.5 * sp[19];
        A[5] = 0.5 * sp[18];
        A[6] = -0.5 * sp[18] + -0.5 * sp[17];
        A[7] = 0.5 * sp[18];
        A[8] = 0.5 * sp[17];
        }

        void tabulate_tensor_poissonL(double* restrict A, const double* w,
                                      const double* c,
                                      const double* restrict coordinate_dofs,
                                      const int* entity_local_index,
                                      const int* cell_orientation,
                                      void* custom_data)
        {
        // Precomputed values of basis functions and precomputations
        // FE* dimensions: [entities][points][dofs]
        // PI* dimensions: [entities][dofs][dofs] or [entities][dofs]
        // PM* dimensions: [entities][dofs][dofs]
        static const double FE4_C0_D01_Q1[1][1][2] = { { { -1.0, 1.0 } } };
        // Unstructured piecewise computations
        const double J_c0 = coordinate_dofs[0] * FE4_C0_D01_Q1[0][0][0]
          + coordinate_dofs[3] * FE4_C0_D01_Q1[0][0][1];
        const double J_c3 = coordinate_dofs[1] * FE4_C0_D01_Q1[0][0][0]
          + coordinate_dofs[7] * FE4_C0_D01_Q1[0][0][1];
        const double J_c1 = coordinate_dofs[0] * FE4_C0_D01_Q1[0][0][0]
          + coordinate_dofs[6] * FE4_C0_D01_Q1[0][0][1];
        const double J_c2 = coordinate_dofs[1] * FE4_C0_D01_Q1[0][0][0]
          + coordinate_dofs[4] * FE4_C0_D01_Q1[0][0][1];
        double sp[4];
        sp[0] = J_c0 * J_c3;
        sp[1] = J_c1 * J_c2;
        sp[2] = sp[0] + -1 * sp[1];
        sp[3] = fabs(sp[2]);
        A[0] = 0.1666666666666667 * sp[3];
        A[1] = 0.1666666666666667 * sp[3];
        A[2] = 0.1666666666666667 * sp[3];
        }
        """,
        )
        ffibuilder.cdef(
            """
        void tabulate_tensor_poissonA(double* restrict A, const double* w,
                                    const double* c,
                                    const double* restrict coordinate_dofs,
                                    const int* entity_local_index,
                                    const int* cell_orientation,
                                    void* custom_data);
        void tabulate_tensor_poissonL(double* restrict A, const double* w,
                                    const double* c,
                                    const double* restrict coordinate_dofs,
                                    const int* entity_local_index,
                                    const int* cell_orientation,
                                    void* custom_data);
        """
        )

        ffibuilder.compile(verbose=True)

    mesh.comm.Barrier()
    from _cffi_kernelA import ffi, lib

    cells = np.arange(mesh.topology.index_map(mesh.topology.dim).size_local, dtype=np.int32)

    ptrA = ffi.cast("intptr_t", ffi.addressof(lib, "tabulate_tensor_poissonA"))
    active_coeffs = np.array([], dtype=np.int8)
    integrals = {IntegralType.cell: [(0, ptrA, cells, active_coeffs)]}
    a = Form(
        _cpp.fem.Form_float64(
            [V._cpp_object, V._cpp_object], integrals, [], [], False, [], mesh=mesh._cpp_object
        )
    )

    ptrL = ffi.cast("intptr_t", ffi.addressof(lib, "tabulate_tensor_poissonL"))
    integrals = {IntegralType.cell: [(0, ptrL, cells, active_coeffs)]}
    L = Form(
        _cpp.fem.Form_float64([V._cpp_object], integrals, [], [], False, [], mesh=mesh._cpp_object)
    )
    A = fem.assemble_matrix(a)
    A.scatter_reverse()
    assert np.isclose(np.sqrt(A.squared_norm()), 56.124860801609124)

    b = fem.assemble_vector(L)
    b.scatter_reverse(la.InsertMode.add)
    assert np.isclose(la.norm(b), 0.0739710713711999)



# The first step is to define the variational problem at hand.
#
# We are interested in solving for a discrete vector field in three
# dimensions, so first we need the appropriate finite element space and
# trial and test functions on this space:

from basix.ufl import element
from ufl import (
    Coefficient,
    Constant,
    FunctionSpace,
    Identity,
    Mesh,
    TestFunction,
    TrialFunction,
    derivative,
    det,
    diff,
    ds,
    dx,
    grad,
    inner,
    ln,
    tr,
    variable,
)

# Function spaces
e = element("Lagrange", "tetrahedron", 1, shape=(3,))
mesh = Mesh(e)
V = FunctionSpace(mesh, e)

# Trial and test functions
du = TrialFunction(V)  # Incremental displacement
v = TestFunction(V)  # Test function

# Note that `element` with `shape=(3,)` creates a finite element space
# of vector fields.
#
# Next, we will be needing functions for the boundary source `B`, the
# traction `T` and the displacement solution itself `u`:

# Functions
u = Coefficient(V)  # Displacement from previous iteration
B = Constant(mesh, shape=(3,))  # Body force per unit volume
T = Constant(mesh, shape=(3,))  # Traction force on the boundary

# Now, we can define the kinematic quantities involved in the model:

# Kinematics
d = len(u)
I = Identity(d)  # Identity tensor  # noqa: E741
F = variable(I + grad(u))  # Deformation gradient
C = F.T * F  # Right Cauchy-Green tensor

# Invariants of deformation tensors
Ic = tr(C)
J = det(F)

# Before defining the energy density and thus the total potential
# energy, it only remains to specify constants for the elasticity
# parameters:

# Elasticity parameters
E = 10.0
nu = 0.3
mu = E / (2 * (1 + nu))
lmbda = E * nu / ((1 + nu) * (1 - 2 * nu))

# Both the first variation of the potential energy, and the Jacobian of
# the variation, can be automatically computed by a call to
# `derivative`:

# Stored strain energy density (compressible neo-Hookean model)
psi = (mu / 2) * (Ic - 3) - mu * ln(J) + (lmbda / 2) * (ln(J)) ** 2

# Total potential energy
Pi = psi * dx - inner(B, u) * dx - inner(T, u) * ds

# First variation of Pi (directional derivative about u in the direction
# of v)
F_form = derivative(Pi, u, v)

# Compute Jacobian of F
J_form = derivative(F_form, u, du)

# Compute Cauchy stress
sigma = (1 / J) * diff(psi, F) * F.T

forms = [F_form, J_form]
elements = [e]
expressions = [(sigma, [[0.25, 0.25, 0.25]])]



# Copyright (C) 2021-2022 Jørgen S. Dokken and Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Support functions for plotting"""

import functools

import numpy as np

from dolfinx import cpp as _cpp
from dolfinx import fem, mesh

# NOTE: This dictionary and the below function that uses it should be
# revised when pyvista improves rendering of 'arbitrary' Lagrange
# elements, i.e. for the VTK cell types that define a shape but allow
# arbitrary degree geometry. See
# https://github.com/pyvista/pyvista/issues/947.
#
# Cell types can be found at
# https://vtk.org/doc/nightly/html/vtkCellType_8h_source.html
_first_order_vtk = {
    mesh.CellType.interval: 3,
    mesh.CellType.triangle: 5,
    mesh.CellType.quadrilateral: 9,
    mesh.CellType.tetrahedron: 10,
    mesh.CellType.hexahedron: 12,
}


@functools.singledispatch
def vtk_mesh(msh: mesh.Mesh, dim: int | None = None, entities=None):
    """Create vtk mesh topology data for mesh entities of a given
    dimension. The vertex indices in the returned topology array are the
    indices for the associated entry in the mesh geometry.

    Args:
        mesh: Mesh to extract data from.
        dim: Topological dimension of entities to extract.
        entities: Entities to extract. Extract all if ``None``.

    Returns:
        Topology, type for each cell, and geometry in VTK-ready format.

    """
    if dim is None:
        dim = msh.topology.dim

    cell_type = _cpp.mesh.cell_entity_type(msh.topology.cell_type, dim, 0)
    if cell_type == mesh.CellType.prism:
        raise RuntimeError("Plotting of prism meshes not supported")

    # Use all local cells if not supplied
    if entities is None:
        entities = np.arange(msh.topology.index_map(dim).size_local, dtype=np.int32)

    geometry_entities = _cpp.mesh.entities_to_geometry(msh._cpp_object, dim, entities, False)

    num_nodes_per_cell = geometry_entities.shape[1]
    map_vtk = np.argsort(_cpp.io.perm_vtk(cell_type, num_nodes_per_cell))
    vtk_topology = geometry_entities[:, map_vtk]

    # Create mesh topology
    topology = np.empty((len(entities), num_nodes_per_cell + 1), dtype=np.int32)
    topology[:, 0] = num_nodes_per_cell
    topology[:, 1:] = vtk_topology

    # Array holding the cell type (shape) for each cell
    vtk_type = _cpp.io.get_vtk_cell_type(cell_type, dim)
    cell_types = np.full(len(entities), vtk_type)

    return topology.reshape(-1), cell_types, msh.geometry.x


@vtk_mesh.register
def _(V: fem.FunctionSpace, entities=None):  # type: ignore
    """Creates a VTK mesh topology (topology array and array of cell
    types) that is based on the degree-of-freedom coordinates.

    This function supports visualisation when the degree of the finite
    element space is different from the geometric degree of the mesh.

    Note:
        This function supports Lagrange elements (continuous and
        discontinuous) only.

    Args:
        V: Mesh to extract data from.
        entities: Entities to extract. Extract all if ``None``.

    Returns:
        Topology, type for each cell, and geometry in VTK-ready format.

    """
    if V.ufl_element().family_name not in [
        "Discontinuous Lagrange",
        "Lagrange",
        "DQ",
        "Q",
        "DP",
        "P",
    ]:
        raise RuntimeError(
            "Can only create meshes from continuous or discontinuous Lagrange"
            "spaces, not {V.ufl_element().family_name}."
        )

    degree = V.ufl_element().degree
    if degree == 0:
        raise RuntimeError("Cannot create topology from cellwise constants.")

    # Use all local cells if not supplied
    msh = V.mesh
    tdim = msh.topology.dim
    if entities is None:
        entities = range(msh.topology.index_map(tdim).size_local)

    dofmap = V.dofmap
    num_dofs_per_cell = V.dofmap.dof_layout.num_dofs
    cell_type = msh.topology.cell_type
    perm = np.argsort(_cpp.io.perm_vtk(cell_type, num_dofs_per_cell))

    vtk_type = (
        _first_order_vtk[cell_type] if degree == 1 else _cpp.io.get_vtk_cell_type(cell_type, tdim)
    )
    cell_types = np.full(len(entities), vtk_type)

    topology = np.zeros((len(entities), num_dofs_per_cell + 1), dtype=np.int32)
    topology[:, 0] = num_dofs_per_cell
    dofmap_ = dofmap.list
    topology[:, 1:] = dofmap_[entities][:, perm]
    return topology.reshape(1, -1)[0], cell_types, V.tabulate_dof_coordinates()



# Copyright (C) 2017-2018 Chris N. Richardson and Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Just-in-time (JIT) compilation using FFCx"""

import functools
import json
import os
import sys
from pathlib import Path

from mpi4py import MPI

import ffcx
import ffcx.codegeneration.jit
import ufl

__all__ = ["ffcx_jit", "get_options", "mpi_jit_decorator"]

DOLFINX_DEFAULT_JIT_OPTIONS = {
    "cache_dir": (
        os.getenv("XDG_CACHE_HOME", default=Path.home().joinpath(".cache")) / Path("fenics"),
        "Path for storing DOLFINx JIT cache. "
        "Default prefix ~/.cache/ can be changed using XDG_CACHE_HOME environment variable.",
    ),
    "cffi_debug": (False, "CFFI debug mode"),
    "cffi_verbose": (False, "CFFI verbose mode"),
    "cffi_libraries": (None, "Extra libraries to link"),
    "timeout": (10, "Timeout for JIT compilation"),
}

if sys.platform.startswith("win32"):
    DOLFINX_DEFAULT_JIT_OPTIONS["cffi_extra_compile_args"] = (
        ["-O2"],
        "Extra C compiler arguments to pass to CFFI",
    )
else:
    DOLFINX_DEFAULT_JIT_OPTIONS["cffi_extra_compile_args"] = (
        ["-O2", "-g0"],
        "Extra C compiler arguments to pass to CFFI",
    )


def mpi_jit_decorator(local_jit, *args, **kwargs):
    """A decorator for jit compilation.

    Use this function as a decorator to any jit compiler function. In a
    parallel run, this function will first call the jit compilation
    function on the first process. When this is done, and the module is
    in the cache, it will call the jit compiler on the remaining
    processes, which will then use the cached module.
    """

    @functools.wraps(local_jit)
    def mpi_jit(comm, *args, **kwargs):
        # Just call JIT compiler when running in serial
        if comm.size == 1:
            return local_jit(*args, **kwargs)

        # Default status (0 == ok, 1 == fail)
        status = 0

        # Compile first on process 0
        root = comm.rank == 0
        if root:
            try:
                output = local_jit(*args, **kwargs)
            except Exception as e:
                status = 1
                error_msg = str(e)

        # TODO: This would have lower overhead if using the dijitso.jit
        # features to inject a waiting callback instead of waiting out
        # here. That approach allows all processes to first look in the
        # cache, introducing a barrier only on cache miss. There's also
        # a sketch in dijitso of how to make only one process per
        # physical cache directory do the compilation.

        # Wait for the compiling process to finish and get status
        # TODO: Would be better to broadcast the status from root but
        # this works.
        global_status = comm.allreduce(status, op=MPI.MAX)
        if global_status == 0:
            # Success, call jit on all other processes (this should just
            # read the cache)
            if not root:
                output = local_jit(*args, **kwargs)
        else:
            # Fail simultaneously on all processes, to allow catching
            # the error without deadlock
            if not root:
                error_msg = "Compilation failed on root node."
            raise RuntimeError(f"Failed just-in-time compilation of form: {error_msg}")

        return output

    # Return the decorated jit function
    return mpi_jit


@functools.cache
def _load_options():
    """Loads options from JSON files."""
    user_config_file = os.getenv("XDG_CONFIG_HOME", default=Path.home().joinpath(".config")) / Path(
        "dolfinx", "dolfinx_jit_options.json"
    )
    try:
        with open(user_config_file) as f:
            user_options = json.load(f)
    except FileNotFoundError:
        user_options = dict()

    pwd_config_file = Path.cwd().joinpath("dolfinx_jit_options.json")
    try:
        with open(pwd_config_file) as f:
            pwd_options = json.load(f)
    except FileNotFoundError:
        pwd_options = dict()

    return (user_options, pwd_options)


def get_options(priority_options: dict | None = None) -> dict:
    """Return a copy of the merged JIT option values for DOLFINx.

    Args:
        priority_options: Take priority over all other option values
            (see notes).

    Returns:
        dict: Merged option values.

    Note:
        See :func:`ffcx_jit` for user facing documentation.

    """
    options = dict()
    for param, (value, _) in DOLFINX_DEFAULT_JIT_OPTIONS.items():
        options[param] = value

    # NOTE: _load_options uses functools.lru_cache
    user_options, pwd_options = _load_options()

    options.update(user_options)
    options.update(pwd_options)
    if priority_options is not None:
        options.update(priority_options)

    options["cache_dir"] = Path(str(options["cache_dir"])).expanduser()

    return options


@mpi_jit_decorator
def ffcx_jit(
    ufl_object, form_compiler_options: dict | None = None, jit_options: dict | None = None
):
    """Compile UFL object with FFCx and CFFI.

    Args:
        ufl_object: Object to compile, e.g. ``ufl.Form``.
        form_compiler_options: Options used in FFCx compilation of
            this form. Execute ``print(ffcx.options.FFCX_DEFAULT_OPTIONS)``
            to see all available options. Takes priority over all other
            option values.
        jit_options: Options used in CFFI JIT compilation of C code
            generated by FFCx. Execute
            ``print(dolfinx.jit.DOLFINX_DEFAULT_JIT_OPTIONS)`` to see all
            available options. Takes priority over all other option values.

    Returns:
        A tuple containing the compiled object, module and a tuple
        containing the header and implementation code.

    Priority ordering of options controlling DOLFINx JIT
    compilation from highest to lowest is:

    -  ``jit_options`` (API, recommended),
    -  ``$PWD/dolfinx_jit_options.json`` (local options),
    -  ``$XDG_CONFIG_HOME/dolfinx/dolfinx_jit_options.json``
       (user options),
    -  default ``DOLFINX_DEFAULT_JIT_OPTIONS`` dictionary in
       :mod:`dolfinx.jit`.

    Priority ordering of options controlling FFCx from highest to
    lowest is:

    -  ``form_compiler_options`` (API, recommended),
    -  ``$PWD/ffcx_options.json`` (local options),
    -  ``$XDG_CONFIG_HOME/ffcx/ffcx_options.json`` (user options),
    -  ``FFCX_DEFAULT_OPTIONS`` in ``ffcx.options``.

    ``$XDG_CONFIG_HOME`` is ``~/.config/`` if the environment variable is
    not set.

    The contents of the ``dolfinx_options.json`` files are cached
    on the first call. Subsequent calls to this function use this
    cache.

    For example, ``dolfinx_jit_options.json`` could contain:

    Example::

        { "cffi_extra_compile_args": ["-O2", "-march=native" ],
          "cffi_verbose": True }

    """
    p_ffcx = ffcx.get_options(form_compiler_options)
    p_jit = get_options(jit_options)

    # Switch on type and compile, returning cffi object
    if isinstance(ufl_object, ufl.Form):
        r = ffcx.codegeneration.jit.compile_forms([ufl_object], options=p_ffcx, **p_jit)
    elif isinstance(ufl_object, tuple) and isinstance(ufl_object[0], ufl.core.expr.Expr):
        r = ffcx.codegeneration.jit.compile_expressions([ufl_object], options=p_ffcx, **p_jit)
    else:
        raise TypeError(type(ufl_object))

    return (r[0][0], r[1], r[2])



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.6
# ---

# # Solving PDEs with different scalar (float) types
#
# ```{admonition} Download sources
# :class: download
# * {download}`Python script <./demo_types.py>`
# * {download}`Jupyter notebook <./demo_types.ipynb>`
# ```

# This demo shows:
#
# - How to solve problems using different scalar types, .e.g. single or
#   double precision, or complex numbers
# - Interfacing with [SciPy](https://scipy.org/) sparse linear algebra
#   functionality


# +
import sys
from pathlib import Path

from mpi4py import MPI

import numpy as np
import scipy.sparse
import scipy.sparse.linalg

import ufl
from dolfinx import fem, la, mesh, plot

# -

# SciPy solvers do not support MPI, so all computations will be
# performed on a single MPI rank

# +
comm = MPI.COMM_SELF
# -

# We create an output directory for storing results and figures

out_folder = Path("out_types")
out_folder.mkdir(parents=True, exist_ok=True)

# Create a function that solves the Poisson equation using different
# precision float and complex scalar types for the finite element
# solution.


def display_scalar(u, name, filter=np.real):
    """Plot the solution using pyvista"""
    try:
        import pyvista

        cells, types, x = plot.vtk_mesh(u.function_space)
        grid = pyvista.UnstructuredGrid(cells, types, x)
        grid.point_data["u"] = filter(u.x.array)
        grid.set_active_scalars("u")
        plotter = pyvista.Plotter()
        plotter.add_mesh(grid, show_edges=True)
        plotter.add_mesh(grid.warp_by_scalar())
        plotter.add_title(f"{name}: real" if filter is np.real else f"{name}: imag")
        if pyvista.OFF_SCREEN:
            plotter.screenshot(out_folder / f"u_{'real' if filter is np.real else 'imag'}.png")
        else:
            plotter.show()
    except ModuleNotFoundError:
        print("'pyvista' is required to visualise the solution")


def display_vector(u, name, filter=np.real):
    """Plot the solution using pyvista"""
    try:
        import pyvista

        V = u.function_space
        cells, types, x = plot.vtk_mesh(V)
        grid = pyvista.UnstructuredGrid(cells, types, x)
        bs = V.dofmap.index_map_bs
        grid.point_data["u"] = filter(np.insert(u.x.array.reshape(x.shape[0], bs), bs, 0, axis=1))
        plotter = pyvista.Plotter()
        plotter.add_mesh(grid.warp_by_scalar(), show_edges=True)
        plotter.add_title(f"{name}: real" if filter is np.real else f"{name}: imag")
        if pyvista.OFF_SCREEN:
            plotter.screenshot(out_folder / f"u_{'real' if filter is np.real else 'imag'}.png")
        else:
            plotter.show()
    except ModuleNotFoundError:
        print("'pyvista' is required to visualise the solution")


def poisson(dtype):
    """Poisson problem solver

    Args:
        dtype: Scalar type to use.


    """

    # Create a mesh and locate facets by a geometric condition
    msh = mesh.create_rectangle(
        comm=comm,
        points=((0.0, 0.0), (2.0, 1.0)),
        n=(32, 16),
        cell_type=mesh.CellType.triangle,
        dtype=np.real(dtype(0)).dtype,
    )
    facets = mesh.locate_entities_boundary(
        msh, dim=1, marker=lambda x: np.isclose(x[0], 0.0) | np.isclose(x[0], 2.0)
    )

    # Define a variational problem.
    V = fem.functionspace(msh, ("Lagrange", 1))
    u, v = ufl.TrialFunction(V), ufl.TestFunction(V)
    x = ufl.SpatialCoordinate(msh)
    fr = 10 * ufl.exp(-((x[0] - 0.5) ** 2 + (x[1] - 0.5) ** 2) / 0.02)
    fc = ufl.sin(2 * np.pi * x[0]) + 10 * ufl.sin(4 * np.pi * x[1]) * 1j
    gr = ufl.sin(5 * x[0])
    gc = ufl.sin(5 * x[0]) * 1j
    a = ufl.inner(ufl.grad(u), ufl.grad(v)) * ufl.dx
    L = ufl.inner(fr + fc, v) * ufl.dx + ufl.inner(gr + gc, v) * ufl.ds

    # In preparation for constructing Dirichlet boundary conditions, locate
    # facets on the constrained boundary and the corresponding
    # degrees-of-freedom.
    dofs = fem.locate_dofs_topological(V=V, entity_dim=1, entities=facets)

    # Process forms. This will compile the forms for the requested type.
    a0 = fem.form(a, dtype=dtype)
    if np.issubdtype(dtype, np.complexfloating):
        L0 = fem.form(L, dtype=dtype)
    else:
        L0 = fem.form(ufl.replace(L, {fc: 0, gc: 0}), dtype=dtype)

    # Create a Dirichlet boundary condition
    bc = fem.dirichletbc(value=dtype(0), dofs=dofs, V=V)

    # Assemble forms
    A = fem.assemble_matrix(a0, [bc])
    A.scatter_reverse()
    b = fem.assemble_vector(L0)
    fem.apply_lifting(b.array, [a0], bcs=[[bc]])
    b.scatter_reverse(la.InsertMode.add)
    bc.set(b.array)

    # Create a SciPy CSR  matrix that shares data with A and solve
    As = A.to_scipy()
    uh = fem.Function(V, dtype=dtype)
    uh.x.array[:] = scipy.sparse.linalg.spsolve(As, b.array)

    return uh


# Create a function that solves the linearised elasticity equation using
# different precision float and complex scalar types for the finite
# element solution.


def elasticity(dtype) -> fem.Function:
    """Linearised elasticity problem solver."""

    # Create a mesh and locate facets by a geometric condition
    msh = mesh.create_rectangle(
        comm=comm,
        points=((0.0, 0.0), (2.0, 1.0)),
        n=(32, 16),
        cell_type=mesh.CellType.triangle,
        dtype=np.real(dtype(0)).dtype,
    )
    facets = mesh.locate_entities_boundary(
        msh, dim=1, marker=lambda x: np.isclose(x[0], 0.0) | np.isclose(x[0], 2.0)
    )

    # Define the variational problem.
    gdim = msh.geometry.dim
    V = fem.functionspace(msh, ("Lagrange", 1, (gdim,)))
    ω, ρ = 300.0, 10.0
    x = ufl.SpatialCoordinate(msh)
    f = ufl.as_vector((ρ * ω**2 * x[0], ρ * ω**2 * x[1]))

    E, ν = 1.0e6, 0.3
    μ, λ = E / (2.0 * (1.0 + ν)), E * ν / ((1.0 + ν) * (1.0 - 2.0 * ν))

    def σ(v):
        """Return an expression for the stress σ given a displacement
        field"""
        return 2.0 * μ * ufl.sym(ufl.grad(v)) + λ * ufl.tr(ufl.sym(ufl.grad(v))) * ufl.Identity(
            len(v)
        )

    u, v = ufl.TrialFunction(V), ufl.TestFunction(V)
    a = ufl.inner(σ(u), ufl.grad(v)) * ufl.dx
    L = ufl.inner(f, v) * ufl.dx

    dofs = fem.locate_dofs_topological(V=V, entity_dim=1, entities=facets)

    # Process forms. This will compile the forms for the requested type.
    a0, L0 = fem.form(a, dtype=dtype), fem.form(L, dtype=dtype)

    # Create a Dirichlet boundary condition
    bc = fem.dirichletbc(np.zeros(2, dtype=dtype), dofs, V=V)

    # Assemble forms (CSR matrix)
    A = fem.assemble_matrix(a0, [bc])
    A.scatter_reverse()

    b = fem.assemble_vector(L0)
    fem.apply_lifting(b.array, [a0], bcs=[[bc]])
    b.scatter_reverse(la.InsertMode.add)
    bc.set(b.array)

    # Create a SciPy CSR  matrix that shares data with A and solve
    As = A.to_scipy()
    uh = fem.Function(V, dtype=dtype)
    uh.x.array[:] = scipy.sparse.linalg.spsolve(As, b.array)

    return uh


# Solve problems for different types


uh = poisson(dtype=np.float32)
uh = poisson(dtype=np.float64)
if not sys.platform.startswith("win32"):
    uh = poisson(dtype=np.complex64)
    uh = poisson(dtype=np.complex128)
display_scalar(uh, "poisson", np.real)
display_scalar(uh, "poisson", np.imag)


uh = elasticity(dtype=np.float32)
uh = elasticity(dtype=np.float64)
if not sys.platform.startswith("win32"):
    uh = elasticity(dtype=np.complex64)
    uh = elasticity(dtype=np.complex128)
display_vector(uh, "elasticity", np.real)



# Copyright (C) 2024 Matthew W. Scroggs Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from mpi4py import MPI

import numpy as np
import pytest

import basix
import basix.ufl
import ufl
from dolfinx import default_real_type, la
from dolfinx.fem import (
    Function,
    apply_lifting,
    assemble_matrix,
    assemble_scalar,
    assemble_vector,
    dirichletbc,
    form,
    functionspace,
    locate_dofs_topological,
)
from dolfinx.mesh import CellType, create_unit_cube, create_unit_square, exterior_facet_indices
from ufl import SpatialCoordinate, TestFunction, TrialFunction, div, dx, grad, inner


def run_scalar_test(V, degree, dtype, cg_solver, rtol=None):
    mesh = V.mesh
    u, v = TrialFunction(V), TestFunction(V)
    a = inner(grad(u), grad(v)) * dx

    # Get quadrature degree for bilinear form integrand (ignores effect of non-affine map)
    a = inner(grad(u), grad(v)) * dx(metadata={"quadrature_degree": -1})
    a.integrals()[0].metadata()["quadrature_degree"] = (
        ufl.algorithms.estimate_total_polynomial_degree(a)
    )
    a = form(a, dtype=dtype)

    # Source term
    x = SpatialCoordinate(mesh)
    u_exact = x[1] ** degree
    f = -div(grad(u_exact))

    # Set quadrature degree for linear form integrand (ignores effect of non-affine map)
    L = inner(f, v) * dx(metadata={"quadrature_degree": -1})
    L.integrals()[0].metadata()["quadrature_degree"] = (
        ufl.algorithms.estimate_total_polynomial_degree(L)
    )
    L = form(L, dtype=dtype)

    u_bc = Function(V, dtype=dtype)
    u_bc.interpolate(lambda x: x[1] ** degree)

    # Create Dirichlet boundary condition
    facetdim = mesh.topology.dim - 1
    mesh.topology.create_connectivity(facetdim, mesh.topology.dim)
    bndry_facets = exterior_facet_indices(mesh.topology)
    bdofs = locate_dofs_topological(V, facetdim, bndry_facets)
    bc = dirichletbc(u_bc, bdofs)

    b = assemble_vector(L)
    apply_lifting(b.array, [a], bcs=[[bc]])
    b.scatter_reverse(la.InsertMode.add)
    bc.set(b.array)

    a = form(a, dtype=dtype)
    A = assemble_matrix(a, bcs=[bc])
    A.scatter_reverse()

    uh = Function(V, dtype=dtype)
    cg_solver(mesh.comm, A, b, uh.x, rtol=rtol)
    uh.x.scatter_forward()

    M = (u_exact - uh) ** 2 * dx
    M = form(M, dtype=dtype)
    error = mesh.comm.allreduce(assemble_scalar(M), op=MPI.SUM)

    eps = np.sqrt(np.finfo(dtype).eps)
    assert np.isclose(error, 0, atol=eps)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("degree", range(1, 6))
def test_basix_element_wrapper(degree, dtype, cg_solver):
    ufl_element = basix.ufl.element(
        basix.ElementFamily.P,
        basix.CellType.triangle,
        degree,
        basix.LagrangeVariant.gll_isaac,
        dtype=dtype,
    )
    mesh = create_unit_square(MPI.COMM_WORLD, 10, 10, dtype=dtype)
    V = functionspace(mesh, ufl_element)
    run_scalar_test(V, degree, dtype, cg_solver)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_custom_element_triangle_degree1(dtype, cg_solver):
    wcoeffs = np.eye(3)
    z = np.zeros((0, 2))
    x = [
        [np.array([[0.0, 0.0]]), np.array([[1.0, 0.0]]), np.array([[0.0, 1.0]])],
        [z, z, z],
        [z],
        [],
    ]
    z = np.zeros((0, 1, 0, 1))
    M = [[np.array([[[[1.0]]]]), np.array([[[[1.0]]]]), np.array([[[[1.0]]]])], [z, z, z], [z], []]
    ufl_element = basix.ufl.custom_element(
        basix.CellType.triangle,
        [],
        wcoeffs,
        x,
        M,
        0,
        basix.MapType.identity,
        basix.SobolevSpace.H1,
        False,
        1,
        1,
        dtype=dtype,
    )
    mesh = create_unit_square(MPI.COMM_WORLD, 10, 10, dtype=dtype)
    V = functionspace(mesh, ufl_element)
    run_scalar_test(V, 1, dtype, cg_solver)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_custom_element_triangle_degree4(dtype, cg_solver):
    wcoeffs = np.eye(15)
    x = [
        [np.array([[0.0, 0.0]]), np.array([[1.0, 0.0]]), np.array([[0.0, 1.0]])],
        [
            np.array([[0.75, 0.25], [0.5, 0.5], [0.25, 0.75]]),
            np.array([[0.0, 0.25], [0.0, 0.5], [0.0, 0.75]]),
            np.array([[0.25, 0.0], [0.5, 0.0], [0.75, 0.0]]),
        ],
        [np.array([[0.25, 0.25], [0.5, 0.25], [0.25, 0.5]])],
        [],
    ]
    id = np.array([[[[1.0], [0.0], [0.0]]], [[[0.0], [1.0], [0.0]]], [[[0.0], [0.0], [1.0]]]])
    M = [
        [np.array([[[[1.0]]]]), np.array([[[[1.0]]]]), np.array([[[[1.0]]]])],
        [id, id, id],
        [id],
        [],
    ]

    ufl_element = basix.ufl.custom_element(
        basix.CellType.triangle,
        [],
        wcoeffs,
        x,
        M,
        0,
        basix.MapType.identity,
        basix.SobolevSpace.H1,
        False,
        4,
        4,
        dtype=dtype,
    )
    mesh = create_unit_square(MPI.COMM_WORLD, 10, 10, dtype=dtype)
    V = functionspace(mesh, ufl_element)
    run_scalar_test(V, 4, dtype, cg_solver)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_custom_element_triangle_degree4_integral(dtype, cg_solver):
    pts, wts = basix.make_quadrature(basix.CellType.interval, 10)
    tab = basix.create_element(
        basix.ElementFamily.P, basix.CellType.interval, 2, dtype=default_real_type
    ).tabulate(0, pts)[0, :, :, 0]
    wcoeffs = np.eye(15)
    x = [
        [np.array([[0.0, 0.0]]), np.array([[1.0, 0.0]]), np.array([[0.0, 1.0]])],
        [
            np.array([[1.0 - p[0], p[0]] for p in pts]),
            np.array([[0.0, p[0]] for p in pts]),
            np.array([[p[0], 0.0] for p in pts]),
        ],
        [np.array([[0.25, 0.25], [0.5, 0.25], [0.25, 0.5]])],
        [],
    ]

    assert pts.shape[0] != 3
    quadrature_mat = np.zeros([3, 1, pts.shape[0], 1])
    for dof in range(3):
        for p in range(pts.shape[0]):
            quadrature_mat[dof, 0, p, 0] = wts[p] * tab[p, dof]

    M = [
        [np.array([[[[1.0]]]]), np.array([[[[1.0]]]]), np.array([[[[1.0]]]])],
        [quadrature_mat, quadrature_mat, quadrature_mat],
        [np.array([[[[1.0], [0.0], [0.0]]], [[[0.0], [1.0], [0.0]]], [[[0.0], [0.0], [1.0]]]])],
        [],
    ]

    ufl_element = basix.ufl.custom_element(
        basix.CellType.triangle,
        [],
        wcoeffs,
        x,
        M,
        0,
        basix.MapType.identity,
        basix.SobolevSpace.H1,
        False,
        4,
        4,
        dtype=dtype,
    )
    mesh = create_unit_square(MPI.COMM_WORLD, 10, 10, dtype=dtype)
    V = functionspace(mesh, ufl_element)
    run_scalar_test(V, 4, dtype, cg_solver, rtol=1e-5)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_custom_element_quadrilateral_degree1(dtype, cg_solver):
    wcoeffs = np.eye(4)
    z = np.zeros((0, 2))
    x = [
        [
            np.array([[0.0, 0.0]]),
            np.array([[1.0, 0.0]]),
            np.array([[0.0, 1.0]]),
            np.array([[1.0, 1.0]]),
        ],
        [z, z, z, z],
        [z],
        [],
    ]
    z = np.zeros((0, 1, 0, 1))
    M = [
        [
            np.array([[[[1.0]]]]),
            np.array([[[[1.0]]]]),
            np.array([[[[1.0]]]]),
            np.array([[[[1.0]]]]),
        ],
        [z, z, z, z],
        [z],
        [],
    ]
    ufl_element = basix.ufl.custom_element(
        basix.CellType.quadrilateral,
        [],
        wcoeffs,
        x,
        M,
        0,
        basix.MapType.identity,
        basix.SobolevSpace.H1,
        False,
        1,
        1,
        dtype=dtype,
    )
    mesh = create_unit_square(MPI.COMM_WORLD, 10, 10, CellType.quadrilateral, dtype=dtype)
    V = functionspace(mesh, ufl_element)
    run_scalar_test(V, 1, dtype, cg_solver)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize(
    "cell_type",
    [CellType.triangle, CellType.quadrilateral, CellType.tetrahedron, CellType.hexahedron],
)
@pytest.mark.parametrize(
    "element_family",
    [
        basix.ElementFamily.N1E,
        basix.ElementFamily.N2E,
        basix.ElementFamily.RT,
        basix.ElementFamily.BDM,
    ],
)
def test_vector_copy_degree1(cell_type, element_family, dtype):
    if cell_type in [CellType.triangle, CellType.quadrilateral]:
        tdim = 2
        mesh = create_unit_square(MPI.COMM_WORLD, 10, 10, cell_type, dtype=dtype)
    else:
        tdim = 3
        mesh = create_unit_cube(MPI.COMM_WORLD, 5, 5, 5, cell_type, dtype=dtype)

    def func(x):
        return x[:tdim]

    e1 = basix.ufl.element(element_family, getattr(basix.CellType, cell_type.name), 1, dtype=dtype)
    e2 = basix.ufl.custom_element(
        e1._element.cell_type,
        e1._element.value_shape,
        e1._element.wcoeffs,
        e1._element.x,
        e1._element.M,
        0,
        e1._element.map_type,
        e1._element.sobolev_space,
        e1._element.discontinuous,
        e1._element.embedded_subdegree,
        e1._element.embedded_superdegree,
        dtype=dtype,
    )

    space1 = functionspace(mesh, e1)
    space2 = functionspace(mesh, e2)

    f1 = Function(space1, dtype=dtype)
    f2 = Function(space2, dtype=dtype)
    f1.interpolate(func)
    f2.interpolate(func)

    diff = f1 - f2
    error = assemble_scalar(form(ufl.inner(diff, diff) * ufl.dx, dtype=dtype))
    assert np.isclose(error, 0)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize(
    "cell_type",
    [CellType.triangle, CellType.quadrilateral, CellType.tetrahedron, CellType.hexahedron],
)
@pytest.mark.parametrize("element_family", [basix.ElementFamily.P, basix.ElementFamily.serendipity])
def test_scalar_copy_degree1(cell_type, element_family, dtype):
    if element_family == basix.ElementFamily.serendipity and cell_type in [
        CellType.triangle,
        CellType.tetrahedron,
    ]:
        pytest.xfail("Serendipity elements cannot be created on simplices")

    if cell_type in [CellType.triangle, CellType.quadrilateral]:
        mesh = create_unit_square(MPI.COMM_WORLD, 10, 10, cell_type, dtype=dtype)
    else:
        mesh = create_unit_cube(MPI.COMM_WORLD, 5, 5, 5, cell_type, dtype=dtype)

    def func(x):
        return x[0]

    e1 = basix.ufl.element(element_family, getattr(basix.CellType, cell_type.name), 1, dtype=dtype)
    e2 = basix.ufl.custom_element(
        e1._element.cell_type,
        e1._element.value_shape,
        e1._element.wcoeffs,
        e1._element.x,
        e1._element.M,
        0,
        e1._element.map_type,
        e1._element.sobolev_space,
        e1._element.discontinuous,
        e1._element.embedded_subdegree,
        e1._element.embedded_superdegree,
        dtype=dtype,
    )

    space1 = functionspace(mesh, e1)
    space2 = functionspace(mesh, e2)

    f1 = Function(space1, dtype=dtype)
    f2 = Function(space2, dtype=dtype)
    f1.interpolate(func)
    f2.interpolate(func)

    diff = f1 - f2
    error = assemble_scalar(form(ufl.inner(diff, diff) * ufl.dx, dtype=dtype))
    assert np.isclose(error, 0)



# Copyright (C) 2009-2020 Garth N. Wells, Matthew W. Scroggs and Jorgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Test that interpolation is done correctly"""

import random

from mpi4py import MPI

import numpy as np
import pytest

import basix
import ufl
from basix.ufl import blocked_element, custom_element, element, enriched_element, mixed_element
from dolfinx import default_real_type, default_scalar_type
from dolfinx.fem import (
    Expression,
    Function,
    assemble_scalar,
    create_interpolation_data,
    form,
    functionspace,
)
from dolfinx.geometry import bb_tree, compute_collisions_points
from dolfinx.mesh import (
    CellType,
    create_mesh,
    create_rectangle,
    create_submesh,
    create_unit_cube,
    create_unit_square,
    locate_entities,
    locate_entities_boundary,
    meshtags,
)

parametrize_cell_types = pytest.mark.parametrize(
    "cell_type",
    [
        CellType.interval,
        CellType.triangle,
        CellType.tetrahedron,
        CellType.quadrilateral,
        CellType.hexahedron,
    ],
)


def random_point_in_reference(cell_type):
    if cell_type == CellType.interval:
        return (random.random(), 0, 0)
    elif cell_type == CellType.triangle:
        x, y = random.random(), random.random()
        # If point is outside cell, move it back inside
        if x + y > 1:
            x, y = 1 - x, 1 - y
        return (x, y, 0)
    elif cell_type == CellType.tetrahedron:
        x, y, z = random.random(), random.random(), random.random()
        # If point is outside cell, move it back inside
        if x + y > 1:
            x, y = 1 - x, 1 - y
        if y + z > 1:
            y, z = 1 - z, 1 - x - y
        if x + y + z > 1:
            x, z = 1 - x - y, x + y + z - 1
        return (x, y, z)
    elif cell_type == CellType.quadrilateral:
        x, y = random.random(), random.random()
        return (x, y, 0)
    elif cell_type == CellType.hexahedron:
        return (random.random(), random.random(), random.random())


def random_point_in_cell(mesh):
    cell_type = mesh.topology.cell_type
    point = random_point_in_reference(cell_type)

    if cell_type == CellType.interval:
        origin = mesh.geometry.x[0]
        axes = (mesh.geometry.x[1],)
    elif cell_type == CellType.triangle:
        origin = mesh.geometry.x[0]
        axes = (mesh.geometry.x[1], mesh.geometry.x[2])
    elif cell_type == CellType.tetrahedron:
        origin = mesh.geometry.x[0]
        axes = (mesh.geometry.x[1], mesh.geometry.x[2], mesh.geometry.x[3])
    elif cell_type == CellType.quadrilateral:
        origin = mesh.geometry.x[0]
        axes = (mesh.geometry.x[1], mesh.geometry.x[2])
    elif cell_type == CellType.hexahedron:
        origin = mesh.geometry.x[0]
        axes = (mesh.geometry.x[1], mesh.geometry.x[2], mesh.geometry.x[4])

    return tuple(
        origin[i] + sum((axis[i] - origin[i]) * p for axis, p in zip(axes, point)) for i in range(3)
    )


def one_cell_mesh(cell_type):
    if cell_type == CellType.interval:
        points = np.array([[-1.0], [2.0]], dtype=default_real_type)
    if cell_type == CellType.triangle:
        points = np.array([[-1.0, -1.0], [2.0, 0.0], [0.0, 0.5]], dtype=default_real_type)
    elif cell_type == CellType.tetrahedron:
        points = np.array(
            [[-1.0, -1.0, -1.0], [2.0, 0.0, 0.0], [0.0, 0.5, 0.0], [0.0, 0.0, 1.0]],
            dtype=default_real_type,
        )
    elif cell_type == CellType.quadrilateral:
        points = np.array(
            [[-1.0, 0.0], [1.0, 0.0], [-1.0, 1.5], [1.0, 1.5]], dtype=default_real_type
        )
    elif cell_type == CellType.hexahedron:
        points = np.array(
            [
                [-1.0, -0.5, 0.0],
                [1.0, -0.5, 0.0],
                [-1.0, 1.5, 0.0],
                [1.0, 1.5, 0.0],
                [0.0, -0.5, 1.0],
                [1.0, -0.5, 1.0],
                [-1.0, 1.5, 1.0],
                [1.0, 1.5, 1.0],
            ],
            dtype=default_real_type,
        )
    num_points = len(points)

    # Randomly number the points and create the mesh
    order = list(range(num_points))
    random.shuffle(order)
    ordered_points = np.zeros(points.shape, dtype=default_real_type)
    for i, j in enumerate(order):
        ordered_points[j] = points[i]
    cells = np.array([order])
    domain = ufl.Mesh(
        element(
            "Lagrange", cell_type.name, 1, shape=(ordered_points.shape[1],), dtype=default_real_type
        )
    )
    return create_mesh(MPI.COMM_WORLD, cells, domain, ordered_points)


def two_cell_mesh(cell_type):
    if cell_type == CellType.interval:
        points = np.array([[0.0], [1.0], [-1.0]], dtype=default_real_type)
        cells = [[0, 1], [0, 2]]
    if cell_type == CellType.triangle:
        # Define equilateral triangles with area 1
        root = 3**0.25  # 4th root of 3
        points = np.array(
            [[0.0, 0.0], [2 / root, 0.0], [1 / root, root], [1 / root, -root]],
            dtype=default_real_type,
        )
        cells = [[0, 1, 2], [1, 0, 3]]
    elif cell_type == CellType.tetrahedron:
        # Define regular tetrahedra with volume 1
        s = 2**0.5 * 3 ** (1 / 3)  # side length
        points = np.array(
            [
                [0.0, 0.0, 0.0],
                [s, 0.0, 0.0],
                [s / 2, s * np.sqrt(3) / 2, 0.0],
                [s / 2, s / 2 / np.sqrt(3), s * np.sqrt(2 / 3)],
                [s / 2, s / 2 / np.sqrt(3), -s * np.sqrt(2 / 3)],
            ],
            dtype=default_real_type,
        )
        cells = [[0, 1, 2, 3], [0, 2, 1, 4]]
    elif cell_type == CellType.quadrilateral:
        # Define unit quadrilaterals (area 1)
        points = np.array(
            [[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0], [0.0, -1.0], [1.0, -1.0]],
            dtype=default_real_type,
        )
        cells = [[0, 1, 2, 3], [5, 1, 4, 0]]
    elif cell_type == CellType.hexahedron:
        # Define unit hexahedra (volume 1)
        points = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
                [1.0, 0.0, 1.0],
                [0.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
                [0.0, 0.0, -1.0],
                [1.0, 0.0, -1.0],
                [0.0, 1.0, -1.0],
                [1.0, 1.0, -1.0],
            ],
            dtype=default_real_type,
        )
        cells = [[0, 1, 2, 3, 4, 5, 6, 7], [9, 11, 8, 10, 1, 3, 0, 2]]

    domain = ufl.Mesh(
        element("Lagrange", cell_type.name, 1, shape=(points.shape[1],), dtype=default_real_type)
    )
    mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)
    return mesh


def run_scalar_test(V, poly_order):
    """Test that interpolation is correct in a scalar valued space."""
    random.seed(13)
    tdim = V.mesh.topology.dim
    if tdim == 1:

        def f(x):
            return x[0] ** poly_order
    elif tdim == 2:

        def f(x):
            return x[1] ** poly_order + 2 * x[0] ** min(poly_order, 1)
    else:

        def f(x):
            return (
                x[1] ** poly_order + 2 * x[0] ** min(poly_order, 1) - 3 * x[2] ** min(poly_order, 2)
            )

    v = Function(V)
    v.interpolate(f)
    points = [random_point_in_cell(V.mesh) for count in range(5)]
    points = np.asarray(points, dtype=default_real_type)
    cells = [0 for count in range(5)]
    values = v.eval(points, cells)
    for p, val in zip(points, values):
        assert np.allclose(val, f(p), atol=1.0e-5)


def run_vector_test(V, poly_order):
    """Test that interpolation is correct in a scalar valued space."""
    random.seed(12)
    tdim = V.mesh.topology.dim

    if tdim == 1:

        def f(x):
            return x[0] ** poly_order
    elif tdim == 2:

        def f(x):
            return (x[1] ** min(poly_order, 1), 2 * x[0] ** poly_order)
    else:

        def f(x):
            return (
                x[1] ** min(poly_order, 1),
                2 * x[0] ** poly_order,
                3 * x[2] ** min(poly_order, 2),
            )

    v = Function(V)
    v.interpolate(f)
    points = [random_point_in_cell(V.mesh) for count in range(5)]
    cells = [0 for count in range(5)]
    values = v.eval(points, cells)
    for p, val in zip(points, values):
        assert np.allclose(val, f(p), atol=1.0e-5)


@pytest.mark.skip_in_parallel
@parametrize_cell_types
@pytest.mark.parametrize("order", range(1, 5))
def test_Lagrange_interpolation(cell_type, order):
    """Test that interpolation is correct in a function space"""
    mesh = one_cell_mesh(cell_type)
    V = functionspace(mesh, ("Lagrange", order))
    run_scalar_test(V, order)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize(
    "cell_type", [CellType.interval, CellType.quadrilateral, CellType.hexahedron]
)
@pytest.mark.parametrize("order", range(1, 5))
def test_serendipity_interpolation(cell_type, order):
    """Test that interpolation is correct in a function space"""
    mesh = one_cell_mesh(cell_type)
    V = functionspace(mesh, ("S", order))
    run_scalar_test(V, order)


@pytest.mark.skip_in_parallel
@parametrize_cell_types
@pytest.mark.parametrize("order", range(1, 5))
def test_vector_interpolation(cell_type, order):
    """Test that interpolation is correct in a blocked (vector) function space."""
    mesh = one_cell_mesh(cell_type)
    gdim = mesh.geometry.dim
    V = functionspace(mesh, ("Lagrange", order, (gdim,)))
    run_vector_test(V, order)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("cell_type", [CellType.triangle, CellType.tetrahedron])
@pytest.mark.parametrize("order", range(1, 5))
def test_N1curl_interpolation(cell_type, order):
    random.seed(8)
    mesh = one_cell_mesh(cell_type)
    V = functionspace(mesh, ("Nedelec 1st kind H(curl)", order))
    run_vector_test(V, order - 1)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("cell_type", [CellType.triangle])
@pytest.mark.parametrize("order", [1, 2])
def test_N2curl_interpolation(cell_type, order):
    mesh = one_cell_mesh(cell_type)
    V = functionspace(mesh, ("Nedelec 2nd kind H(curl)", order))
    run_vector_test(V, order)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("cell_type", [CellType.quadrilateral])
@pytest.mark.parametrize("order", range(1, 5))
def test_RTCE_interpolation(cell_type, order):
    random.seed(8)
    mesh = one_cell_mesh(cell_type)
    V = functionspace(mesh, ("RTCE", order))
    run_vector_test(V, order - 1)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("cell_type", [CellType.hexahedron])
@pytest.mark.parametrize("order", range(1, 5))
def test_NCE_interpolation(cell_type, order):
    random.seed(8)
    mesh = one_cell_mesh(cell_type)
    V = functionspace(mesh, ("NCE", order))
    run_vector_test(V, order - 1)


def test_mixed_sub_interpolation():
    """Test interpolation of sub-functions"""
    mesh = create_unit_cube(MPI.COMM_WORLD, 3, 3, 3)

    def f(x):
        return np.vstack((10 + x[0], -10 - x[1], 25 + x[0]))

    P2 = element(
        "Lagrange", mesh.basix_cell(), 2, shape=(mesh.geometry.dim,), dtype=default_real_type
    )
    P1 = element("Lagrange", mesh.basix_cell(), 1, dtype=default_real_type)
    for i, P in enumerate((mixed_element([P2, P1]), mixed_element([P1, P2]))):
        W = functionspace(mesh, P)
        U = Function(W)
        U.sub(i).interpolate(f)

        # Same element
        V = functionspace(mesh, P2)
        u, v = Function(V), Function(V)
        u.interpolate(U.sub(i))
        v.interpolate(f)
        assert np.allclose(u.x.array, v.x.array)

        # Same map, different elements
        gdim = mesh.geometry.dim
        V = functionspace(mesh, ("Lagrange", 1, (gdim,)))
        u, v = Function(V), Function(V)
        u.interpolate(U.sub(i))
        v.interpolate(f)
        assert np.allclose(u.x.array, v.x.array)

        # Different maps (0)
        V = functionspace(mesh, ("N1curl", 1))
        u, v = Function(V), Function(V)
        u.interpolate(U.sub(i))
        v.interpolate(f)
        atol = 5 * np.finfo(u.x.array.dtype).resolution
        assert np.allclose(u.x.array, v.x.array, atol=atol)

        # Different maps (1)
        V = functionspace(mesh, ("RT", 2))
        u, v = Function(V), Function(V)
        u.interpolate(U.sub(i))
        v.interpolate(f)
        atol = 5 * np.finfo(u.x.array.dtype).resolution
        assert np.allclose(u.x.array, v.x.array, atol=atol)

        # Test with wrong shape
        V0 = functionspace(mesh, P.sub_elements[0])
        V1 = functionspace(mesh, P.sub_elements[1])
        v0, v1 = Function(V0), Function(V1)
        with pytest.raises(RuntimeError):
            v0.interpolate(U.sub(1))
        with pytest.raises(RuntimeError):
            v1.interpolate(U.sub(0))


@pytest.mark.skip_in_parallel
def test_mixed_interpolation():
    """Test that mixed interpolation raised an exception."""
    mesh = one_cell_mesh(CellType.triangle)
    A = element("Lagrange", mesh.basix_cell(), 1, dtype=default_real_type)
    B = element(
        "Lagrange", mesh.basix_cell(), 1, shape=(mesh.geometry.dim,), dtype=default_real_type
    )
    v = Function(functionspace(mesh, mixed_element([A, B])))
    with pytest.raises(RuntimeError):
        v.interpolate(lambda x: (x[1], 2 * x[0], 3 * x[1]))


@pytest.mark.parametrize("order1", [2, 3, 4])
@pytest.mark.parametrize("order2", [2, 3, 4])
def test_interpolation_nedelec(order1, order2):
    mesh = create_unit_cube(MPI.COMM_WORLD, 2, 2, 2)
    V = functionspace(mesh, ("N1curl", order1))
    V1 = functionspace(mesh, ("N1curl", order2))
    u, v = Function(V), Function(V1)

    # The expression "lambda x: x" is contained in the N1curl function
    # space for order > 1
    u.interpolate(lambda x: x)
    v.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - v, u - v) * ufl.dx)) == pytest.approx(0, abs=1.0e-10)

    # The target expression is also contained in N2curl space of any
    # order
    V2 = functionspace(mesh, ("N2curl", 1))
    w = Function(V2)
    w.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - w, u - w) * ufl.dx)) == pytest.approx(0, abs=1.0e-10)


@pytest.mark.parametrize("tdim", [2, 3])
@pytest.mark.parametrize("order", [1, 2, 3])
def test_interpolation_dg_to_n1curl(tdim, order):
    if tdim == 2:
        mesh = create_unit_square(MPI.COMM_WORLD, 5, 5)
    else:
        mesh = create_unit_cube(MPI.COMM_WORLD, 2, 2, 2)
    V = functionspace(mesh, ("DG", order, (tdim,)))
    V1 = functionspace(mesh, ("N1curl", order + 1))
    u, v = Function(V), Function(V1)
    u.interpolate(lambda x: x[:tdim] ** order)
    v.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - v, u - v) * ufl.dx)) == pytest.approx(0.0, abs=1.0e-8)


@pytest.mark.parametrize("tdim", [2, 3])
@pytest.mark.parametrize("order", [1, 2, 3])
def test_interpolation_n1curl_to_dg(tdim, order):
    if tdim == 2:
        mesh = create_unit_square(MPI.COMM_WORLD, 5, 5)
    else:
        mesh = create_unit_cube(MPI.COMM_WORLD, 2, 2, 2)
    V = functionspace(mesh, ("N1curl", order + 1))
    V1 = functionspace(mesh, ("DG", order, (tdim,)))
    u, v = Function(V), Function(V1)
    u.interpolate(lambda x: x[:tdim] ** order)
    v.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - v, u - v) * ufl.dx)) == pytest.approx(0.0, abs=1e-10)


@pytest.mark.parametrize("tdim", [2, 3])
@pytest.mark.parametrize("order", [1, 2, 3])
def test_interpolation_n2curl_to_bdm(tdim, order):
    if tdim == 2:
        mesh = create_unit_square(MPI.COMM_WORLD, 5, 5)
    else:
        mesh = create_unit_cube(MPI.COMM_WORLD, 2, 2, 2)
    V = functionspace(mesh, ("N2curl", order))
    V1 = functionspace(mesh, ("BDM", order))
    u, v = Function(V), Function(V1)
    u.interpolate(lambda x: x[:tdim] ** order)
    v.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - v, u - v) * ufl.dx)) == pytest.approx(
        0.0, abs=1.0e-10
    )


@pytest.mark.parametrize("order1", [1, 2, 3, 4, 5])
@pytest.mark.parametrize("order2", [1, 2, 3])
def test_interpolation_p2p(order1, order2):
    mesh = create_unit_cube(MPI.COMM_WORLD, 2, 2, 2)
    V = functionspace(mesh, ("Lagrange", order1))
    V1 = functionspace(mesh, ("Lagrange", order2))
    u, v = Function(V), Function(V1)
    u.interpolate(lambda x: x[0])
    v.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - v, u - v) * ufl.dx)) == pytest.approx(0.0, abs=1e-10)

    DG = functionspace(mesh, ("DG", order2))
    w = Function(DG)
    w.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - w, u - w) * ufl.dx)) == pytest.approx(0.0, abs=1e-10)


@pytest.mark.parametrize("order1", [1, 2, 3])
@pytest.mark.parametrize("order2", [1, 2])
def test_interpolation_vector_elements(order1, order2):
    mesh = create_unit_cube(MPI.COMM_WORLD, 2, 2, 2)
    gdim = mesh.geometry.dim
    V = functionspace(mesh, ("Lagrange", order1, (gdim,)))
    V1 = functionspace(mesh, ("Lagrange", order2, (gdim,)))
    u, v = Function(V), Function(V1)
    u.interpolate(lambda x: x)
    v.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - v, u - v) * ufl.dx)) == pytest.approx(0)

    DG = functionspace(mesh, ("DG", order2, (gdim,)))
    w = Function(DG)
    w.interpolate(u)
    assert assemble_scalar(form(ufl.inner(u - w, u - w) * ufl.dx)) == pytest.approx(0)


@pytest.mark.skip_in_parallel
def test_interpolation_non_affine():
    points = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 2, 0],
            [1, 2, 0],
            [0, 0, 3],
            [1, 0, 3],
            [0, 2, 3],
            [1, 2, 3],
            [0.5, 0, 0],
            [0, 1, 0],
            [0, 0, 1.5],
            [1, 1, 0],
            [1, 0, 1.5],
            [0.5, 2, 0],
            [0, 2, 1.5],
            [1, 2, 1.5],
            [0.5, 0, 3],
            [0, 1, 3],
            [1, 1, 3],
            [0.5, 2, 3],
            [0.5, 1, 0],
            [0.5, 0, 1.5],
            [0, 1, 1.5],
            [1, 1, 1.5],
            [0.5, 2, 1.5],
            [0.5, 1, 3],
            [0.5, 1, 1.5],
        ],
        dtype=default_real_type,
    )
    cells = np.array([range(len(points))], dtype=np.int32)
    domain = ufl.Mesh(element("Lagrange", "hexahedron", 2, shape=(3,), dtype=default_real_type))
    mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)
    W = functionspace(mesh, ("NCE", 1))
    V = functionspace(mesh, ("NCE", 2))
    w, v = Function(W), Function(V)
    w.interpolate(lambda x: x)
    v.interpolate(w)
    assert assemble_scalar(form(ufl.inner(w - v, w - v) * ufl.dx)) == pytest.approx(0, abs=1e-10)


@pytest.mark.skip_in_parallel
def test_interpolation_non_affine_nonmatching_maps():
    points = np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [0, 2, 0],
            [1, 2, 0],
            [0, 0, 3],
            [1, 0, 3],
            [0, 2, 3],
            [1, 2, 3],
            [0.5, 0, 0],
            [0, 1, 0],
            [0, 0, 1.5],
            [1, 1, 0],
            [1, 0, 1.5],
            [0.5, 2, 0],
            [0, 2, 1.5],
            [1, 2, 1.5],
            [0.5, 0, 3],
            [0, 1, 3],
            [1, 1, 3],
            [0.5, 2, 3],
            [0.5, 1, 0],
            [0.5, -0.1, 1.5],
            [0, 1, 1.5],
            [1, 1, 1.5],
            [0.5, 2, 1.5],
            [0.5, 1, 3],
            [0.5, 1, 1.5],
        ],
        dtype=default_real_type,
    )
    cells = np.array([range(len(points))], dtype=np.int32)
    domain = ufl.Mesh(element("Lagrange", "hexahedron", 2, shape=(3,), dtype=default_real_type))
    mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)
    gdim = mesh.geometry.dim
    W = functionspace(mesh, ("DG", 1, (gdim,)))
    V = functionspace(mesh, ("NCE", 4))
    w, v = Function(W), Function(V)
    w.interpolate(lambda x: x)
    v.interpolate(w)
    assert assemble_scalar(form(ufl.inner(w - v, w - v) * ufl.dx)) == pytest.approx(0, abs=1e-8)


@pytest.mark.parametrize("order", [2, 3, 4])
@pytest.mark.parametrize("dim", [2, 3])
def test_nedelec_spatial(order, dim):
    if dim == 2:
        mesh = create_unit_square(MPI.COMM_WORLD, 4, 4)
    elif dim == 3:
        mesh = create_unit_cube(MPI.COMM_WORLD, 2, 2, 2)

    V = functionspace(mesh, ("N1curl", order))
    u = Function(V)
    x = ufl.SpatialCoordinate(mesh)

    # The expression (x,y,z) is contained in the N1curl function space
    # order>1
    f_ex = x
    f = Expression(f_ex, V.element.interpolation_points)
    u.interpolate(f)
    assert np.abs(assemble_scalar(form(ufl.inner(u - f_ex, u - f_ex) * ufl.dx))) == pytest.approx(
        0, abs=1e-10
    )

    # The target expression is also contained in N2curl space of any
    # order
    V2 = functionspace(mesh, ("N2curl", 1))
    w = Function(V2)
    f2 = Expression(f_ex, V2.element.interpolation_points)
    w.interpolate(f2)
    assert np.abs(assemble_scalar(form(ufl.inner(w - f_ex, w - f_ex) * ufl.dx))) == pytest.approx(0)


@pytest.mark.parametrize("order", [1, 2, 3, 4])
@pytest.mark.parametrize("dim", [2, 3])
@pytest.mark.parametrize("affine", [True, False])
def test_vector_interpolation_spatial(order, dim, affine):
    if dim == 2:
        ct = CellType.triangle if affine else CellType.quadrilateral
        mesh = create_unit_square(MPI.COMM_WORLD, 3, 4, ct)
    elif dim == 3:
        ct = CellType.tetrahedron if affine else CellType.hexahedron
        mesh = create_unit_cube(MPI.COMM_WORLD, 3, 2, 2, ct)

    V = functionspace(mesh, ("Lagrange", order, (dim,)))
    u = Function(V)
    x = ufl.SpatialCoordinate(mesh)

    # The expression (x,y,z)^n is contained in space
    f = ufl.as_vector([x[i] ** order for i in range(dim)])
    u.interpolate(Expression(f, V.element.interpolation_points))
    assert np.abs(assemble_scalar(form(ufl.inner(u - f, u - f) * ufl.dx))) == pytest.approx(0)


@pytest.mark.parametrize("order", [1, 2, 3, 4])
def test_2D_lagrange_to_curl(order):
    mesh = create_unit_square(MPI.COMM_WORLD, 3, 4)
    V, W = functionspace(mesh, ("N1curl", order)), functionspace(mesh, ("Lagrange", order))
    u, u0 = Function(V), Function(W)
    u0.interpolate(lambda x: -x[1])
    u1 = Function(W)
    u1.interpolate(lambda x: x[0])
    f = ufl.as_vector((u0, u1))
    f_expr = Expression(f, V.element.interpolation_points)
    u.interpolate(f_expr)
    x = ufl.SpatialCoordinate(mesh)
    f_ex = ufl.as_vector((-x[1], x[0]))
    assert np.abs(assemble_scalar(form(ufl.inner(u - f_ex, u - f_ex) * ufl.dx))) == pytest.approx(0)


@pytest.mark.parametrize("order", [2, 3, 4])
def test_de_rahm_2D(order):
    mesh = create_unit_square(MPI.COMM_WORLD, 3, 4)
    W = functionspace(mesh, ("Lagrange", order))
    w = Function(W)
    w.interpolate(lambda x: x[0] + x[0] * x[1] + 2 * x[1] ** 2)
    g = ufl.grad(w)
    Q = functionspace(mesh, ("N2curl", order - 1))
    q = Function(Q)
    q.interpolate(Expression(g, Q.element.interpolation_points))
    x = ufl.SpatialCoordinate(mesh)
    g_ex = ufl.as_vector((1 + x[1], 4 * x[1] + x[0]))
    assert np.abs(assemble_scalar(form(ufl.inner(q - g_ex, q - g_ex) * ufl.dx))) == pytest.approx(
        0, abs=np.sqrt(np.finfo(mesh.geometry.x.dtype).eps)
    )

    V = functionspace(mesh, ("BDM", order - 1))
    v = Function(V)

    def curl2D(u):
        return ufl.as_vector((ufl.Dx(u[1], 0), -ufl.Dx(u[0], 1)))

    v.interpolate(Expression(curl2D(ufl.grad(w)), V.element.interpolation_points))
    h_ex = ufl.as_vector((1, -1))
    assert np.abs(assemble_scalar(form(ufl.inner(v - h_ex, v - h_ex) * ufl.dx))) == pytest.approx(
        0, abs=np.sqrt(np.finfo(mesh.geometry.x.dtype).eps)
    )


@pytest.mark.parametrize("order", [1, 2, 3, 4])
@pytest.mark.parametrize("dim", [2, 3])
@pytest.mark.parametrize("affine", [True, False])
@pytest.mark.parametrize("callable_", [True, False])
def test_interpolate_subset(order, dim, affine, callable_):
    if dim == 2:
        ct = CellType.triangle if affine else CellType.quadrilateral
        mesh = create_unit_square(MPI.COMM_WORLD, 3, 4, ct)
    elif dim == 3:
        ct = CellType.tetrahedron if affine else CellType.hexahedron
        mesh = create_unit_cube(MPI.COMM_WORLD, 3, 2, 2, ct)

    V = functionspace(mesh, ("DG", order))
    u = Function(V)

    cells = locate_entities(mesh, mesh.topology.dim, lambda x: x[1] <= 0.5 + 1e-10)
    num_local_cells = mesh.topology.index_map(mesh.topology.dim).size_local
    cells_local = cells[cells < num_local_cells]
    x = ufl.SpatialCoordinate(mesh)
    f = x[1] ** order
    if not callable_:
        expr = Expression(f, V.element.interpolation_points)
        u.interpolate(expr, cells_local)
    else:
        u.interpolate(lambda x: x[1] ** order, cells_local)
    mt = meshtags(mesh, mesh.topology.dim, cells_local, np.ones(cells_local.size, dtype=np.int32))
    dx = ufl.Measure("dx", domain=mesh, subdomain_data=mt)
    assert np.abs(form(assemble_scalar(form(ufl.inner(u - f, u - f) * dx(1))))) == pytest.approx(0)
    integral = mesh.comm.allreduce(assemble_scalar(form(u * dx)), op=MPI.SUM)
    assert integral == pytest.approx(1 / (order + 1) * 0.5 ** (order + 1), abs=1.0e-6)


def test_interpolate_callable():
    """Test interpolation with callables"""
    numba = pytest.importorskip("numba")
    mesh = create_unit_square(MPI.COMM_WORLD, 2, 1)
    V = functionspace(mesh, ("Lagrange", 2))
    u0, u1 = Function(V), Function(V)

    @numba.njit
    def f(x):
        return x[0]

    u0.interpolate(lambda x: x[0])
    u1.interpolate(f)
    assert np.allclose(u0.x.array, u1.x.array)
    with pytest.raises(RuntimeError):
        u0.interpolate(lambda x: np.vstack([x[0], x[1]]))


@pytest.mark.parametrize("bound", [1.5, 0.5])
def test_interpolate_callable_subset(bound):
    """Test interpolation on subsets with callables"""
    mesh = create_unit_square(MPI.COMM_WORLD, 3, 4)
    cells = locate_entities(mesh, mesh.topology.dim, lambda x: x[1] <= bound + 1e-10)
    num_local_cells = mesh.topology.index_map(mesh.topology.dim).size_local
    cells_local = cells[cells < num_local_cells]
    V = functionspace(mesh, ("DG", 2))
    u0, u1 = Function(V), Function(V)
    x = ufl.SpatialCoordinate(mesh)
    f = x[0]
    expr = Expression(f, V.element.interpolation_points)
    u0.interpolate(lambda x: x[0], cells_local)
    u1.interpolate(expr, cells_local)
    assert np.allclose(u0.x.array, u1.x.array, rtol=1.0e-6, atol=1.0e-6)


@pytest.mark.parametrize(
    "scalar_element",
    [
        element("P", "triangle", 1, dtype=default_real_type),
        element("P", "triangle", 2, dtype=default_real_type),
        element("P", "triangle", 3, dtype=default_real_type),
        element("Q", "quadrilateral", 1, dtype=default_real_type),
        element("Q", "quadrilateral", 2, dtype=default_real_type),
        element("Q", "quadrilateral", 3, dtype=default_real_type),
        element("S", "quadrilateral", 1, dtype=default_real_type),
        element("S", "quadrilateral", 2, dtype=default_real_type),
        element("S", "quadrilateral", 3, dtype=default_real_type),
        enriched_element(
            [
                element("P", "triangle", 1, dtype=default_real_type),
                element("Bubble", "triangle", 3, dtype=default_real_type),
            ]
        ),
        enriched_element(
            [
                element("P", "quadrilateral", 1, dtype=default_real_type),
                element("Bubble", "quadrilateral", 2, dtype=default_real_type),
            ]
        ),
    ],
)
def test_vector_element_interpolation(scalar_element):
    """Test interpolation into a range of vector elements."""
    mesh = create_unit_square(
        MPI.COMM_WORLD, 10, 10, getattr(CellType, scalar_element.cell.cellname())
    )
    V = functionspace(mesh, blocked_element(scalar_element, shape=(2,)))
    u = Function(V)
    u.interpolate(lambda x: (x[0], x[1]))
    u2 = Function(V)
    u2.sub(0).interpolate(lambda x: x[0])
    u2.sub(1).interpolate(lambda x: x[1])
    assert np.allclose(u2.x.array, u.x.array)


def test_custom_vector_element():
    """Test interpolation into an element with a value size that uses an identity map."""
    mesh = create_unit_square(MPI.COMM_WORLD, 10, 10)
    wcoeffs = np.eye(6)
    x = [[], [], [], []]
    x[0].append(np.array([[0.0, 0.0]]))
    x[0].append(np.array([[1.0, 0.0]]))
    x[0].append(np.array([[0.0, 1.0]]))
    for _ in range(3):
        x[1].append(np.zeros((0, 2)))
    x[2].append(np.zeros((0, 2)))
    M = [[], [], [], []]
    for _ in range(3):
        M[0].append(np.array([[[[1.0]], [[0.0]]], [[[0.0]], [[1.0]]]]))
    for _ in range(3):
        M[1].append(np.zeros((0, 2, 0, 1)))
    M[2].append(np.zeros((0, 2, 0, 1)))
    e = custom_element(
        basix.CellType.triangle,
        [2],
        wcoeffs,
        x,
        M,
        0,
        basix.MapType.identity,
        basix.SobolevSpace.H1,
        False,
        1,
        1,
        dtype=default_real_type,
    )

    V = functionspace(mesh, e)
    gdim = mesh.geometry.dim
    W = functionspace(mesh, ("Lagrange", 1, (gdim,)))
    v = Function(V)
    w = Function(W)
    v.interpolate(lambda x: (x[0], x[1]))
    w.interpolate(lambda x: (x[0], x[1]))
    assert np.abs(assemble_scalar(form(ufl.inner(v - w, v - w) * ufl.dx))) == pytest.approx(0)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("cell_type", [CellType.triangle, CellType.tetrahedron])
@pytest.mark.parametrize("order", range(1, 5))
def test_mixed_interpolation_permuting(cell_type, order):
    random.seed(8)
    mesh = two_cell_mesh(cell_type)

    def g(x):
        return np.sin(x[1]) + 2 * x[0]

    x = ufl.SpatialCoordinate(mesh)
    dgdy = ufl.cos(x[1])

    curl_el = element("N1curl", mesh.basix_cell(), 1, dtype=default_real_type)
    vlag_el = element(
        "Lagrange", mesh.basix_cell(), 1, shape=(mesh.geometry.dim,), dtype=default_real_type
    )
    lagr_el = element("Lagrange", mesh.basix_cell(), order, dtype=default_real_type)

    V = functionspace(mesh, mixed_element([curl_el, lagr_el]))
    Eb_m = Function(V)
    Eb_m.sub(1).interpolate(g)
    diff = Eb_m[2].dx(1) - dgdy
    error = assemble_scalar(form(ufl.dot(diff, diff) * ufl.dx))

    V = functionspace(mesh, mixed_element([vlag_el, lagr_el]))
    Eb_m = Function(V)
    Eb_m.sub(1).interpolate(g)
    diff = Eb_m[2].dx(1) - dgdy
    assert assemble_scalar(form(ufl.dot(diff, diff) * ufl.dx)) == pytest.approx(error)


@pytest.mark.parametrize("xtype", [np.float64])
@pytest.mark.parametrize("cell_type0", [CellType.hexahedron, CellType.tetrahedron])
@pytest.mark.parametrize("cell_type1", [CellType.triangle, CellType.quadrilateral])
def test_nonmatching_mesh_interpolation(xtype, cell_type0, cell_type1):
    mesh0 = create_unit_cube(MPI.COMM_WORLD, 5, 6, 7, cell_type=cell_type0, dtype=xtype)
    mesh1 = create_unit_square(MPI.COMM_WORLD, 5, 4, cell_type=cell_type1, dtype=xtype)

    def f(x):
        return (7 * x[1], 3 * x[0], x[2] + 0.4)

    el0 = element("Lagrange", mesh0.basix_cell(), 1, shape=(3,), dtype=xtype)
    V0 = functionspace(mesh0, el0)
    el1 = element("Lagrange", mesh1.basix_cell(), 1, shape=(3,), dtype=xtype)
    V1 = functionspace(mesh1, el1)

    # Interpolate on 3D mesh
    u0 = Function(V0, dtype=xtype)
    u0.interpolate(f)
    u0.x.scatter_forward()
    padding = 1e-14

    # Check that both interfaces of create nonmatching meshes interpolation data returns the same
    fine_mesh_cell_map = mesh1.topology.index_map(mesh1.topology.dim)
    num_cells_on_proc = fine_mesh_cell_map.size_local + fine_mesh_cell_map.num_ghosts
    cells = np.arange(num_cells_on_proc, dtype=np.int32)
    interpolation_data = create_interpolation_data(V1, V0, cells, padding=padding)

    # Interpolate 3D->2D
    u1 = Function(V1, dtype=xtype)

    u1.interpolate_nonmatching(u0, cells, interpolation_data=interpolation_data)
    u1.x.scatter_forward()

    # Exact interpolation on 2D mesh
    u1_ex = Function(V1, dtype=xtype)
    u1_ex.interpolate(f)
    u1_ex.x.scatter_forward()

    assert np.allclose(
        u1_ex.x.array,
        u1.x.array,
        rtol=np.sqrt(np.finfo(xtype).eps),
        atol=np.sqrt(np.finfo(xtype).eps),
    )

    # Interpolate 2D->3D
    cell_map0 = mesh0.topology.index_map(mesh0.topology.dim)
    num_cells_on_proc = cell_map0.size_local + cell_map0.num_ghosts
    cells0 = np.arange(num_cells_on_proc, dtype=np.int32)
    interpolation_data1 = create_interpolation_data(V0, V1, cells0, padding=padding)
    u0_2 = Function(V0, dtype=xtype)
    u0_2.interpolate_nonmatching(u1, cells0, interpolation_data1)

    # Check that function values over facets of 3D mesh of the twice
    # interpolated property is preserved
    def locate_bottom_facets(x):
        return np.isclose(x[2], 0)

    facets = locate_entities_boundary(mesh0, mesh0.topology.dim - 1, locate_bottom_facets)
    facet_tag = meshtags(
        mesh0, mesh0.topology.dim - 1, facets, np.full(len(facets), 1, dtype=np.int32)
    )
    residual = ufl.inner(u0 - u0_2, u0 - u0_2) * ufl.ds(
        domain=mesh0, subdomain_data=facet_tag, subdomain_id=1
    )
    assert np.isclose(assemble_scalar(form(residual, dtype=xtype)), 0)


@pytest.mark.parametrize("xtype", [np.float64])
def test_nonmatching_mesh_single_cell_overlap_interpolation(xtype):
    # mesh2 is contained by a single cell of mesh1. Here we test
    # interpolation when *not* every process has data to communicate

    # Test interpolation from mesh1 to mesh2
    n_mesh1 = 2
    mesh1 = create_rectangle(
        MPI.COMM_WORLD,
        [[0.0, 0.0], [1.0, 1.0]],
        [n_mesh1, n_mesh1],
        cell_type=CellType.quadrilateral,
        dtype=xtype,
    )

    n_mesh2 = 2
    p0_mesh2 = 1.0 / n_mesh1
    mesh2 = create_rectangle(
        MPI.COMM_WORLD,
        [[0.0, 0.0], [p0_mesh2, p0_mesh2]],
        [n_mesh2, n_mesh2],
        cell_type=CellType.triangle,
        dtype=xtype,
    )

    u1 = Function(functionspace(mesh1, ("Lagrange", 1)), name="u1", dtype=xtype)
    u2 = Function(functionspace(mesh2, ("Lagrange", 1)), name="u2", dtype=xtype)

    def f_test1(x):
        return 1.0 - x[0] * x[1]

    u1.interpolate(f_test1)
    u1.x.scatter_forward()
    padding = 1e-14
    cell_map2 = mesh2.topology.index_map(mesh2.topology.dim)
    num_cells2 = cell_map2.size_local + cell_map2.num_ghosts
    cells2 = np.arange(num_cells2, dtype=np.int32)
    u1_2_u2_nmm_data = create_interpolation_data(
        u2.function_space, u1.function_space, cells2, padding=padding
    )

    u2.interpolate_nonmatching(u1, cells2, interpolation_data=u1_2_u2_nmm_data)
    u2.x.scatter_forward()

    # interpolate f which is exactly represented on the element
    u2_exact = Function(u2.function_space, dtype=xtype)
    u2_exact.interpolate(f_test1)
    u2_exact.x.scatter_forward()

    l2_error = assemble_scalar(form((u2 - u2_exact) ** 2 * ufl.dx, dtype=xtype))
    assert np.isclose(l2_error, 0.0, rtol=np.finfo(xtype).eps, atol=np.finfo(xtype).eps)

    # Test interpolation from mesh2 to mesh1
    def f_test2(x):
        return x[0] * x[1]

    u1.x.array[:] = 0.0
    u1.x.scatter_forward()

    u2.interpolate(f_test2)
    u2.x.scatter_forward()
    padding = 1e-14
    cell_map1 = mesh1.topology.index_map(mesh1.topology.dim)
    num_cells1 = cell_map1.size_local + cell_map1.num_ghosts
    cells1 = np.arange(num_cells1, dtype=np.int32)
    u2_2_u1_nmm_data = create_interpolation_data(
        u1.function_space, u2.function_space, cells1, padding=padding
    )

    u1.interpolate_nonmatching(u2, cells1, interpolation_data=u2_2_u1_nmm_data)
    u1.x.scatter_forward()

    u1_exact = Function(u1.function_space, dtype=xtype)
    u1_exact.interpolate(f_test2)
    u1_exact.x.scatter_forward()

    # Find the single cell in mesh1 which is overlapped by mesh2
    tree1 = bb_tree(mesh1, mesh1.topology.dim, padding=0.0)
    cells_overlapped1 = compute_collisions_points(
        tree1, np.array([p0_mesh2, p0_mesh2, 0.0]) / 2
    ).array
    assert cells_overlapped1.shape[0] <= 1

    # Construct the error measure on the overlapped cell
    cell_label = 1
    cts = meshtags(
        mesh1, mesh1.topology.dim, cells_overlapped1, np.full_like(cells_overlapped1, cell_label)
    )
    dx_cell = ufl.Measure("dx", subdomain_data=cts)

    l2_error = assemble_scalar(form((u1 - u1_exact) ** 2 * dx_cell(cell_label), dtype=xtype))
    assert np.isclose(l2_error, 0.0, rtol=np.finfo(xtype).eps, atol=np.finfo(xtype).eps)


def test_submesh_interpolation():
    """Test interpolation of a function between a sub-mesh and its parent mesh."""
    mesh = create_unit_square(MPI.COMM_WORLD, 6, 7)

    def left_locator(x):
        return x[0] <= 0.5 + 1e-14

    def ref_func(x):
        return x[0] + x[1] ** 2

    tdim = mesh.topology.dim
    cells = locate_entities(mesh, tdim, left_locator)
    submesh, entity_map, _, _ = create_submesh(mesh, tdim, cells)

    u0 = Function(functionspace(mesh, ("Lagrange", 2)))
    u0.interpolate(ref_func)

    V1 = functionspace(submesh, ("DG", 3))
    u1 = Function(V1)

    smsh_cell_imap = submesh.topology.index_map(tdim)
    smsh_cells = np.arange(smsh_cell_imap.size_local + smsh_cell_imap.num_ghosts)
    parent_cells = entity_map.sub_topology_to_topology(smsh_cells, inverse=False)

    # Interpolate u0 (defined on 'full' mesh) into u0 (defined on
    # 'sub'0mesh)
    u1.interpolate(u0, cells0=parent_cells, cells1=smsh_cells)

    u1_exact = Function(V1)
    u1_exact.interpolate(ref_func)
    atol = 5 * np.finfo(default_scalar_type).resolution
    np.testing.assert_allclose(u1_exact.x.array, u1.x.array, atol=atol)

    # Map from sub to parent
    W = functionspace(mesh, ("DG", 4))
    w = Function(W)

    # Interpolate Function defined on sub-mesh (u1_exact) to the part of
    # a Function on the full mesh (w)
    w.interpolate(u1_exact, cells0=np.arange(len(parent_cells)), cells1=parent_cells)
    w_exact = Function(W)
    w_exact.interpolate(ref_func, cells0=cells)
    np.testing.assert_allclose(w.x.array, w_exact.x.array, atol=atol)


def xtest_submesh_expression_interpolation():
    """Test interpolation of an expression between a submesh and its parent"""
    mesh = create_unit_square(MPI.COMM_WORLD, 10, 8, cell_type=CellType.quadrilateral)

    def left_locator(x):
        return x[0] <= 0.5 + 1e-14

    def ref_func(x):
        return -3 * x[0] ** 2 + x[1] ** 2

    def grad_ref_func(x):
        values = np.zeros((2, x.shape[1]), dtype=default_scalar_type)
        values[0] = -6 * x[0]
        values[1] = 2 * x[1]
        return values

    def modified_grad(x):
        grad = grad_ref_func(x)
        return -0.2 * grad[1], 0.1 * grad[0]

    tdim = mesh.topology.dim
    cells = locate_entities(mesh, tdim, left_locator)
    submesh, sub_to_parent, _, _ = create_submesh(mesh, tdim, cells)

    V = functionspace(mesh, ("Lagrange", 2))
    u = Function(V)
    u.interpolate(ref_func)

    V_sub = functionspace(submesh, ("N2curl", 1))
    u_sub = Function(V_sub)

    parent_expr = Expression(ufl.grad(u), V_sub.element.interpolation_points)

    # Map from parent to sub mesh

    u_sub.interpolate(parent_expr, expr_mesh=mesh, cell_map=sub_to_parent)

    u_sub_exact = Function(V_sub)
    u_sub_exact.interpolate(grad_ref_func)
    atol = 10 * np.finfo(default_scalar_type).resolution
    np.testing.assert_allclose(u_sub_exact.x.array, u_sub.x.array, atol=atol)

    # Map from sub to parent
    W = functionspace(mesh, ("DQ", 1, (mesh.geometry.dim,)))
    w = Function(W)

    cell_imap = mesh.topology.index_map(tdim)
    num_cells = cell_imap.size_local + cell_imap.num_ghosts
    parent_to_sub = np.full(num_cells, -1, dtype=np.int32)
    parent_to_sub[sub_to_parent] = np.arange(len(sub_to_parent))

    # Map exact solution (based on quadrature points) back to parent mesh
    sub_vec = ufl.as_vector((-0.2 * u_sub_exact[1], 0.1 * u_sub_exact[0]))
    sub_expr = Expression(sub_vec, W.element.interpolation_points)

    # Mapping back needs to be restricted to the subset of cells in the submesh
    w.interpolate(sub_expr, cells=sub_to_parent, expr_mesh=submesh, cell_map=parent_to_sub)

    w_exact = Function(W)
    w_exact.interpolate(modified_grad, cells=cells)
    w_exact.x.scatter_forward()
    np.testing.assert_allclose(w.x.array, w_exact.x.array, atol=atol)



# Copyright (C) 2011-2021 Garth N. Wells and Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from pathlib import Path

from mpi4py import MPI

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import ufl
from basix.ufl import element, mixed_element
from dolfinx import default_real_type
from dolfinx.fem import Function, functionspace
from dolfinx.io import VTKFile
from dolfinx.io.utils import cell_perm_vtk  # F401
from dolfinx.mesh import (
    CellType,
    create_mesh,
    create_unit_cube,
    create_unit_interval,
    create_unit_square,
)
from dolfinx.plot import vtk_mesh

cell_types_2D = [CellType.triangle, CellType.quadrilateral]
cell_types_3D = [CellType.tetrahedron, CellType.hexahedron]


def test_save_1d_mesh_subdir(tempdir):
    filename = Path(tempdir, "mesh.pvd")
    mesh = create_unit_interval(MPI.COMM_WORLD, 32)
    with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
        vtk.write_mesh(mesh)
        vtk.write_mesh(mesh, 1)


@pytest.mark.parametrize("cell_type", cell_types_2D)
def test_save_2d_mesh(tempdir, cell_type):
    mesh = create_unit_square(MPI.COMM_WORLD, 32, 32, cell_type=cell_type)
    filename = Path(tempdir, f"mesh_{cell_type.name}.pvd")
    with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
        vtk.write_mesh(mesh, 0.0)
        vtk.write_mesh(mesh, 2.0)


@pytest.mark.parametrize("cell_type", cell_types_3D)
def test_save_3d_mesh(tempdir, cell_type):
    mesh = create_unit_cube(MPI.COMM_WORLD, 8, 8, 8, cell_type=cell_type)
    filename = Path(tempdir, f"mesh_{cell_type.name}.pvd")
    with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
        vtk.write_mesh(mesh, 0.0)
        vtk.write_mesh(mesh, 2.0)


def test_save_1d_scalar(tempdir):
    mesh = create_unit_interval(MPI.COMM_WORLD, 32)
    u = Function(functionspace(mesh, ("Lagrange", 2)))
    u.interpolate(lambda x: x[0])
    filename = Path(tempdir, "u.pvd")
    with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
        vtk.write_function(u, 0.0)


@pytest.mark.parametrize("cell_type", cell_types_2D)
def test_save_2d_scalar(tempdir, cell_type):
    mesh = create_unit_square(MPI.COMM_WORLD, 16, 16, cell_type=cell_type)
    u = Function(functionspace(mesh, ("Lagrange", 2)))
    u.x.array[:] = 1.0

    filename = Path(tempdir, "u.pvd")
    with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
        vtk.write_function(u, 0.0)
        vtk.write_function(u, 1.0)


@pytest.mark.parametrize("cell_type", cell_types_3D)
def test_save_3d_scalar(tempdir, cell_type):
    mesh = create_unit_cube(MPI.COMM_WORLD, 8, 8, 8, cell_type=cell_type)
    u = Function(functionspace(mesh, ("Lagrange", 2)))
    u.x.array[:] = 1.0

    filename = Path(tempdir, "u.pvd")
    with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
        vtk.write_function(u, 0.0)
        vtk.write_function(u, 1.0)


def test_save_1d_vector(tempdir):
    mesh = create_unit_interval(MPI.COMM_WORLD, 32)

    def f(x):
        vals = np.zeros((2, x.shape[1]))
        vals[0] = x[0]
        vals[1] = 2 * x[0] * x[0]
        return vals

    e = element("Lagrange", mesh.basix_cell(), 2, shape=(2,), dtype=default_real_type)
    u = Function(functionspace(mesh, e))
    u.interpolate(f)
    filename = Path(tempdir, "u.pvd")
    with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
        vtk.write_function(u, 0.0)


@pytest.mark.parametrize("cell_type", cell_types_2D)
def test_save_2d_vector(tempdir, cell_type):
    mesh = create_unit_square(MPI.COMM_WORLD, 16, 16, cell_type=cell_type)
    gdim = mesh.geometry.dim
    u = Function(functionspace(mesh, ("Lagrange", 1, (gdim,))))

    def f(x):
        vals = np.zeros((2, x.shape[1]))
        vals[0] = x[0]
        vals[1] = 2 * x[0] * x[1]
        return vals

    u.interpolate(f)
    filename = Path(tempdir, "u.pvd")
    with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
        vtk.write_function(u, 0.0)
        vtk.write_function(u, 1.0)


@pytest.mark.skip_in_parallel
def test_save_2d_vector_CG2(tempdir):
    points = np.array(
        [[0, 0], [1, 0], [1, 2], [0, 2], [1 / 2, 0], [1, 1], [1 / 2, 2], [0, 1], [1 / 2, 1]],
        dtype=default_real_type,
    )
    points = np.array(
        [[0, 0], [1, 0], [0, 2], [0.5, 1], [0, 1], [0.5, 0], [1, 2], [0.5, 2], [1, 1]],
        dtype=default_real_type,
    )
    cells = np.array([[0, 1, 2, 3, 4, 5], [1, 6, 2, 7, 3, 8]])
    domain = ufl.Mesh(element("Lagrange", "triangle", 2, shape=(2,), dtype=default_real_type))
    mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)
    gdim = mesh.geometry.dim
    u = Function(functionspace(mesh, ("Lagrange", 2, (gdim,))))
    u.interpolate(lambda x: np.vstack((x[0], x[1])))
    filename = Path(tempdir, "u.pvd")
    with VTKFile(mesh.comm, filename, "w") as vtk:
        vtk.write_function(u, 0.0)


def test_save_vtk_mixed(tempdir):
    mesh = create_unit_cube(MPI.COMM_WORLD, 3, 3, 3)
    P2 = element(
        "Lagrange", mesh.basix_cell(), 1, shape=(mesh.geometry.dim,), dtype=default_real_type
    )
    P1 = element("Lagrange", mesh.basix_cell(), 1, dtype=default_real_type)
    W = functionspace(mesh, mixed_element([P2, P1]))
    V1 = functionspace(mesh, P1)
    V2 = functionspace(mesh, P2)

    U = Function(W)
    U.sub(0).interpolate(lambda x: np.vstack((x[0], 0.2 * x[1], np.zeros_like(x[0]))))
    U.sub(1).interpolate(lambda x: 0.5 * x[0])

    U1, U2 = Function(V1), Function(V2)
    U1.interpolate(U.sub(1))
    U2.interpolate(U.sub(0))
    U2.name = "u"
    U1.name = "p"

    filename = Path(tempdir, "u.pvd")
    with VTKFile(mesh.comm, filename, "w") as vtk:
        vtk.write_function([U2, U1], 0.0)
    with VTKFile(mesh.comm, filename, "w") as vtk:
        vtk.write_function([U1, U2], 0.0)

    Up = U.sub(1)
    Up.name = "psub"
    with pytest.raises(RuntimeError):
        with VTKFile(mesh.comm, filename, "w") as vtk:
            vtk.write_function([U2, Up, U1], 0)
    with pytest.raises(RuntimeError):
        with VTKFile(mesh.comm, filename, "w") as vtk:
            vtk.write_function([U.sub(i) for i in range(W.num_sub_spaces)], 0)


@pytest.mark.parametrize("cell_type", cell_types_2D)
def test_save_vector_element(tempdir, cell_type):
    mesh = create_unit_square(MPI.COMM_WORLD, 16, 16, cell_type=cell_type)
    u = Function(functionspace(mesh, ("RT", 1)))

    def f(x):
        vals = np.zeros((2, x.shape[1]))
        vals[0] = x[0]
        vals[1] = 2 * x[0] * x[1]
        return vals

    u.interpolate(f)
    filename = Path(tempdir, "u.pvd")
    with pytest.raises(RuntimeError):
        with VTKFile(MPI.COMM_WORLD, filename, "w") as vtk:
            vtk.write_function(u, 0.0)
            vtk.write_function(u, 1.0)


def test_save_vtk_cell_point(tempdir):
    """Test writing cell-wise and point-wise data"""
    mesh = create_unit_cube(MPI.COMM_WORLD, 3, 3, 3)
    P2 = element("Lagrange", mesh.basix_cell(), 1, shape=(3,), dtype=default_real_type)
    P1 = element("Discontinuous Lagrange", mesh.basix_cell(), 0, dtype=default_real_type)

    V2, V1 = functionspace(mesh, P2), functionspace(mesh, P1)
    U2, U1 = Function(V2), Function(V1)
    U2.interpolate(lambda x: np.vstack((x[0], 0.2 * x[1], np.zeros_like(x[0]))))
    U1.interpolate(lambda x: 0.5 * x[0])
    U2.name = "A"
    U1.name = "B"

    filename = Path(tempdir, "u.pvd")
    with VTKFile(mesh.comm, filename, "w") as vtk:
        vtk.write_function([U2, U1], 0.0)
    with VTKFile(mesh.comm, filename, "w") as vtk:
        vtk.write_function((U1, U2), 0.0)


def test_save_1d_tensor(tempdir):
    mesh = create_unit_interval(MPI.COMM_WORLD, 32)
    e = element("Lagrange", mesh.basix_cell(), 2, shape=(2, 2), dtype=default_real_type)
    u = Function(functionspace(mesh, e))
    u.x.array[:] = 1.0
    filename = Path(tempdir, "u.pvd")
    with VTKFile(mesh.comm, filename, "w") as vtk:
        vtk.write_function(u, 0.0)


def test_save_2d_tensor(tempdir):
    mesh = create_unit_square(MPI.COMM_WORLD, 16, 16)
    gdim = mesh.geometry.dim
    u = Function(functionspace(mesh, ("Lagrange", 2, (gdim, gdim))))
    u.x.array[:] = 1.0
    filename = Path(tempdir, "u.pvd")
    with VTKFile(mesh.comm, filename, "w") as vtk:
        vtk.write_function(u, 0.0)
        u.x.array[:] = 2.0
        vtk.write_function(u, 1.0)


def test_save_3d_tensor(tempdir):
    mesh = create_unit_cube(MPI.COMM_WORLD, 8, 8, 8)
    gdim = mesh.geometry.dim
    u = Function(functionspace(mesh, ("Lagrange", 2, (gdim, gdim))))
    u.x.array[:] = 1.0
    filename = Path(tempdir, "u.pvd")
    with VTKFile(mesh.comm, filename, "w") as vtk:
        vtk.write_function(u, 0.0)


def test_triangle_perm_vtk():
    higher_order_triangle_perm = {
        10: np.array([0, 1, 2, 5, 6, 8, 7, 3, 4, 9]),
        15: np.array([0, 1, 2, 6, 7, 8, 11, 10, 9, 3, 4, 5, 12, 13, 14]),
        21: np.array([0, 1, 2, 7, 8, 9, 10, 14, 13, 12, 11, 3, 4, 5, 6, 15, 18, 16, 20, 19, 17]),
        28: np.array(
            [
                0,
                1,
                2,
                8,
                9,
                10,
                11,
                12,
                17,
                16,
                15,
                14,
                13,
                3,
                4,
                5,
                6,
                7,
                18,
                21,
                22,
                19,
                26,
                27,
                23,
                25,
                24,
                20,
            ]
        ),
        36: np.array(
            [
                0,
                1,
                2,
                9,
                10,
                11,
                12,
                13,
                14,
                20,
                19,
                18,
                17,
                16,
                15,
                3,
                4,
                5,
                6,
                7,
                8,
                21,
                24,
                25,
                26,
                22,
                32,
                33,
                34,
                27,
                31,
                35,
                28,
                30,
                29,
                23,
            ]
        ),
        45: np.array(
            [
                0,
                1,
                2,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                23,
                22,
                21,
                20,
                19,
                18,
                17,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                24,
                27,
                28,
                29,
                30,
                25,
                38,
                39,
                42,
                40,
                31,
                37,
                44,
                43,
                32,
                36,
                41,
                33,
                35,
                34,
                26,
            ]
        ),
        55: np.array(
            [
                0,
                1,
                2,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                26,
                25,
                24,
                23,
                22,
                21,
                20,
                19,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                27,
                30,
                31,
                32,
                33,
                34,
                28,
                44,
                45,
                48,
                49,
                46,
                35,
                43,
                53,
                54,
                50,
                36,
                42,
                52,
                51,
                37,
                41,
                47,
                38,
                40,
                39,
                29,
            ]
        ),
    }
    for p_test, v_test in higher_order_triangle_perm.items():
        v = cell_perm_vtk(CellType.triangle, p_test)
        assert_array_equal(v, v_test)


def test_vtk_mesh():
    comm = MPI.COMM_WORLD
    mesh = create_unit_square(comm, 2 * comm.size, 2 * comm.size)
    V = functionspace(mesh, ("Lagrange", 1))
    vtk_mesh(V)



from mpi4py import MPI

import numpy as np

from dolfinx import graph, mesh


def to_adj(cells, dtype):
    cflat = []
    coff = [0]
    for c in cells:
        cflat += c
        cc = coff[-1] + len(c)
        coff += [cc]
    adj = graph.adjacencylist(np.array(cflat, dtype=dtype), np.array(coff, dtype=dtype))
    return adj


def test_dgrsph_1d():
    rank = MPI.COMM_WORLD.Get_rank()
    size = MPI.COMM_WORLD.Get_size()
    n0 = rank * 3
    x = n0 + 3
    if rank == size - 1:
        x = 0
    # Circular chain of interval cells
    cells = [[n0, n0 + 1], [n0 + 1, n0 + 2], [n0 + 2, x]]
    w = mesh.build_dual_graph(
        MPI.COMM_WORLD, mesh.CellType.interval, to_adj(cells, np.int64)._cpp_object
    )
    assert w.num_nodes == 3
    for i in range(w.num_nodes):
        assert len(w.links(i)) == 2



# This demo aims to illustrate how to assemble a matrix with a trial function
# defined on a submesh of co-dimension 0, and a test function defined on the parent mesh
from basix.ufl import element
from ufl import (
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dx,
    inner,
)

cell = "quadrilateral"
coord_element = element("Lagrange", cell, 1, shape=(2,))
mesh = Mesh(coord_element)

# We define the function space and test function on the full mesh
e = element("Lagrange", cell, 1)
V = FunctionSpace(mesh, e)
v = TestFunction(V)

# Next we define the sub-mesh
submesh = Mesh(coord_element)
W = FunctionSpace(submesh, e)
p = TrialFunction(W)

# And finally we define a "mass matrix" on the submesh, with the test function
# of the parent mesh. The integration domain is the parent mesh, but we restrict integration
# to all cells marked with subdomain_id=3, which will indicate what cells of our mesh is part
# of the submesh
a_mixed = inner(p, v) * dx(domain=mesh, subdomain_id=3)

q = TestFunction(W)
a = inner(p, q) * dx(domain=submesh)

forms = [a_mixed, a]



# Copyright (C) 2013 Anders Logg
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from mpi4py import MPI

import numpy as np
import pytest

import basix
import ufl
from basix.ufl import element
from dolfinx.geometry import squared_distance
from dolfinx.mesh import create_mesh, create_unit_interval


@pytest.mark.skip_in_parallel
def test_distance_interval():
    mesh = create_unit_interval(MPI.COMM_SELF, 1)
    d = np.array([-1.0, 0.0, 0.0])
    assert squared_distance(mesh, mesh.topology.dim, np.array([0]), d) == pytest.approx(1.0)
    d = np.array([0.5, 0.0, 0.0])
    assert squared_distance(mesh, mesh.topology.dim, np.array([0]), d) == pytest.approx(0.0)


@pytest.mark.skip_in_parallel
def test_distance_triangle():
    shape, degree = "triangle", 1
    domain = basix.create_element(basix.ElementFamily.P, basix.CellType[shape], degree)
    x = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 0.0]], dtype=np.float64)
    cells = [[0, 1, 2]]
    mesh = create_mesh(MPI.COMM_WORLD, cells, domain, x)
    d = np.array([-1.0, -1.0, 0.0])
    assert squared_distance(mesh, mesh.topology.dim, np.array([0]), d) == pytest.approx(2.0)
    d = np.array([-1.0, 0.5, 0.0])
    assert squared_distance(mesh, mesh.topology.dim, np.array([0]), d) == pytest.approx(1.0)
    d = np.array([0.5, 0.5, 0.0])
    assert squared_distance(mesh, mesh.topology.dim, np.array([0]), d) == pytest.approx(0.0)


@pytest.mark.skip_in_parallel
def test_distance_tetrahedron():
    shape = "tetrahedron"
    degree = 1
    domain = ufl.Mesh(element("Lagrange", shape, degree, shape=(3,), dtype=np.float64))
    x = np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 1.0, 1.0], [1, 1.0, 1]], dtype=np.float64)
    cells = [[0, 1, 2, 3]]
    mesh = create_mesh(MPI.COMM_WORLD, cells, domain, x)
    d = np.array([-1.0, -1.0, -1.0])
    assert squared_distance(mesh, mesh.topology.dim, np.array([0]), d) == pytest.approx(3.0)
    d = np.array([-1.0, 0.5, 0.5])
    assert squared_distance(mesh, mesh.topology.dim, np.array([0]), d) == pytest.approx(1.0)
    d = np.array([0.5, 0.5, 0.5])
    assert squared_distance(mesh, mesh.topology.dim, np.array([0]), d) == pytest.approx(0.0)


# @pytest.mark.skip("volume_entities needs fixing")
# @pytest.mark.parametrize(
#     'mesh', [
#         create_unit_interval(MPI.COMM_WORLD, 18),
#         create_unit_square(MPI.COMM_WORLD, 8, 9, CellType.triangle),
#         create_unit_square(MPI.COMM_WORLD, 8, 9, CellType.quadrilateral),
#         create_unit_cube(MPI.COMM_WORLD, 8, 9, 5, CellType.tetrahedron)
#     ])
# def test_volume_cells(mesh):
#     tdim = mesh.topology.dim
#     map = mesh.topology.index_map(tdim)
#     num_cells = map.size_local
#     v = _cpp.mesh.volume_entities(mesh, range(num_cells), mesh.topology.dim)
#     assert mesh.comm.allreduce(v.sum(), MPI.SUM) == pytest.approx(1.0, rel=1e-9)


# @pytest.mark.skip("volume_entities needs fixing")
# def test_volume_quadrilateralR2():
#     mesh = create_unit_square(MPI.COMM_SELF, 1, 1, CellType.quadrilateral)
#     assert _cpp.mesh.volume_entities(mesh, [0], mesh.topology.dim) == 1.0


# @pytest.mark.skip("volume_entities needs fixing")
# @pytest.mark.parametrize(
#     'x',
#     [[[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [1.0, 1.0, 0.0]],
#      [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [0.0, 1.0, 1.0]]])
# def test_volume_quadrilateralR3(x):
#     cells = [[0, 1, 2, 3]]
#     domain = ufl.Mesh(element("Lagrange", "quadrilateral", 1, shape=(2,)))
#     mesh = create_mesh(MPI.COMM_SELF, cells, domain, x)
#     assert _cpp.mesh.volume_entities(mesh, [0], mesh.topology.dim) == 1.0


# @pytest.mark.skip("volume_entities needs fixing")
# @pytest.mark.parametrize(
#     'scaling',
#     [1e0, 1e-5, 1e-10, 1e-15, 1e-20, 1e-30, 1e5, 1e10, 1e15, 1e20, 1e30])
# def test_volume_quadrilateral_coplanarity_check_1(scaling):
#     with pytest.raises(RuntimeError) as error:
#         # Unit square cell scaled down by 'scaling' and the first vertex
#         # is distorted so that the vertices are clearly non coplanar
#         x = [[scaling, 0.5 * scaling, 0.6 * scaling],
#              [0.0, scaling, 0.0],
#              [0.0, 0.0, scaling],
#              [0.0, scaling, scaling]]
#         cells = [[0, 1, 2, 3]]
#         domain = ufl.Mesh(element("Lagrange", "quadrilateral", 1, shape=(2,)))
#         mesh = create_mesh(MPI.COMM_SELF, cells, domain, x)
#         _cpp.mesh.volume_entities(mesh, [0], mesh.topology.dim)

#     assert "Not coplanar" in str(error.value)


# Test when |p0-p3| is ~ 1 but |p1-p2| is small
# The cell is degenerate when scale is below 1e-17, it is expected to
# fail the test.
# @pytest.mark.skip("volume_entities needs fixing")
# @pytest.mark.parametrize('scaling', [1e0, 1e-5, 1e-10, 1e-15])
# def test_volume_quadrilateral_coplanarity_check_2(scaling):
#     with pytest.raises(RuntimeError) as error:
#         # Unit square cell scaled down by 'scaling' and the first vertex
#         # is distorted so that the vertices are clearly non coplanar
#         x = [[1.0, 0.5, 0.6], [0.0, scaling, 0.0],
#              [0.0, 0.0, scaling], [0.0, 1.0, 1.0]]
#         cells = [[0, 1, 2, 3]]
#         domain = ufl.Mesh(element("Lagrange", "quadrilateral", 1, shape=(2,)))
#         mesh = create_mesh(MPI.COMM_SELF, cells, domain, x)
#         _cpp.mesh.volume_entities(mesh, [0], mesh.topology.dim)

#     assert "Not coplanar" in str(error.value)



# Copyright (C) 2021-2025 Chris Richardson, Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from packaging.version import parse

import dolfinx


def test_version():
    """Test that installed Python version matches C++ version.

    Python DOLFINx follows `major.minor.micro` with the append of `.devx`
    and `.postx` denoting development and postrelease, respectively.

    C++ DOLFINx follows `major.minor.micro` with a development denoted
    `major.minor.micro.0`. postrelease cannot be reflected in C++,
    any changes to the C++ code should bump micro.
    """
    cpp_version = parse(dolfinx.cpp.__version__)
    python_version = parse(dolfinx.__version__)

    assert cpp_version.major == python_version.major
    assert cpp_version.minor == python_version.minor
    assert cpp_version.micro == python_version.micro

    cpp_is_devrelease = True if len(cpp_version.release) == 4 else False
    assert cpp_is_devrelease == python_version.is_devrelease
